#include "YumIO.h"
#include "YumIOC.h"
#include <QMessageBox>
#include <QRegExp>
#include <QEventLoop>
#include <QFile>
#include "ScrollableErrorDialog.h"
#include <qglobal.h>
#include <QThread>
#include <QMutex>
#include <QTimer>
#include "Share.h"
#include <rpm/rpmlib.h>
#include <rpm/rpmcli.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <datetime.h>
#include "stringmacros.h"
#include "globals.h"

#undef DEBUG_COLLECT_DOWNLOAD
#undef DEBUG_COLLECT_DOWNLOAD_COOKED
#undef DEBUG_TS
#undef DEBUG_TS_DESCR
#undef DEBUG_TS_DOWNLOAD
#define DEBUG_TS_DOWNLOAD_COOKED
#undef DEBUG_TS_DETAIL
#undef DEBUG_EXPECTED_EXCEPTIONS
#undef DEBUG_ENVRA
#undef DEBUG_IO
#define DEBUG_DISABLE_CAPTURERS
// To fully disable all I/O interception, also comment out these two lines at
// the end of python/YumIO.py:
//    #sys.stdout = YumIOStdout()
//    #sys.stderr = YumIOStderr()
//    #sys.stdin = YumIOStdin()

// dprintf() is a debug macro that always dumps to stderr, working around
// the capturers.  It does use variadic macros, though, so it requires a
// compiler that supports them (i.e. gcc).
#define dprintf(f, ...)  (fprintf(file_dbg, (f), ## __VA_ARGS__))

#define UNUSED __attribute((unused))

static inline long long  max (long long  l,
                              long long  r)
{
   if (l >= r) {
      return l;
   } else {
      return r;
   }
}

#if PY_VERSION_HEX < 0x02050000
typedef int  Py_ssize_t;
#endif

#if PY_VERSION_HEX >= 0x03000000
#define PY_TO_LONG(obj)  (PyLong_AsLong(obj))
#else
#define PY_TO_LONG(obj)  (PyInt_AsLong(obj))
#endif

#if PY_VERSION_HEX >= 0x03000000
#define PY_IS_STRING(obj)  (PyUnicode_Check(obj))
#else
#define PY_IS_STRING(obj)  (PyString_Check(obj) || PyUnicode_Check(obj))
#endif

//
// Values passed to dnf.yum.rpmtrans.LoggingTransactionDisplay.progress
//
#define PKG_INSTALL          1
#define PKG_DOWNGRADE        2
#define PKG_OBSOLETE         4
#define PKG_UPGRADE          6
#define PKG_REMOVE           8
#define PKG_REINSTALL        9
#define PKG_CLEANUP        101
#define PKG_VERIFY         102
#define PKG_SCRIPTLET      103
#define TRANS_PREPARATION  201
#define TRANS_POST         202

#define IS_INSTALL(action) \
   ((action) == TS_INSTALL || (action) == TS_TRUEINSTALL)
#define IS_UPDATE(action) \
   ((action) == TS_UPDATE || (action) == TS_OBSOLETING)
#define IS_REMOVE(action) \
   ((action) == TS_ERASE)
#define IS_COMPLETING_UPDATE(action) \
   ((action) == TS_UPDATED || (action) == TS_OBSOLETED)

//
// Known yum strings for collect's progress indications
//
#define COLL_GRAB_METADATA  "MD Read"

//
// Known yum strings for available packages
//
#define YUM_DECODE_PRCO_FLAGS(f)  (strcmp((f), "GT") == 0 ? ">"  : \
                                   strcmp((f), "GE") == 0 ? ">=" : \
                                   strcmp((f), "LT") == 0 ? "<"  : \
                                   strcmp((f), "LE") == 0 ? "<=" : \
                                   strcmp((f), "EQ") == 0 ? "="  : \
                                                            (f))

// When using package sizes to estimate the amount of time for a package,
// very small packages still will have overhead time.  So, use a lower bound
// of PKG_MIN as the "size" of those packages when figuring the total
// progress.  (Continue to use the actual size for item progress, though.)
#define PKG_MIN  10240ULL

#define RA_IMAGE(act)                         \
   (act == RA_Grab_Metalink ? "Grab_Meta"   : \
    act == RA_Grab_RepoMD   ? "Grab_RepoMD" : \
    act == RA_Grab_Group    ? "Grab_Group"  : \
    act == RA_Grab          ? "Grab"        : \
    act == RA_Convert       ? "Convert"     : \
    act == RA_Read          ? "Read"        : \
                              "?")

//
// Progress values for collect
//
#define PG_C_INITIALIZE        100
#define PG_C_DOWNLOAD_START    900
#define PG_C_DOWNLOAD        14000
#define PG_C_UPDATES          1000
#define PG_C_INSTALL          1000
#define PG_C_GROUPS           1000

#define PG_C_TOTAL_TEXT           "Initializing: Total Progress"
#define PG_C_INITIALIZE_TEXT      "Preparing"
#define PG_C_DOWNLOAD_START_TEXT  "Reading repositories"
//      PG_C_DOWNLOAD_TEXT        "Downloading repository metadata for ..."
#define PG_C_UPDATES_TEXT         "Collecting updatable packages"
#define PG_C_INSTALL_TEXT         "Collecting installed and available packages"
#define PG_C_GROUPS_TEXT          "Collecting groups"

#define PG_C_TOTAL  (  PG_C_INITIALIZE     \
                     + PG_C_DOWNLOAD_START \
                     + PG_C_DOWNLOAD       \
                     + PG_C_UPDATES        \
                     + PG_C_INSTALL        \
                     + PG_C_GROUPS)

//
// Progress values for collectFileLists
//
#define PG_F_REPOS_START   1000
#define PG_F_REPOS        14000

#define PG_F_TOTAL_TEXT        "Loading file lists: Total Progress"
#define PG_F_REPOS_START_TEXT  "Reading repositories"
//      PG_F_REPOS_TEXT        "Downloading/Reading repository metadata for ..."

#define PG_F_TOTAL  (  PG_F_REPOS_START \
                     + PG_F_REPOS)

//
// Progress values for collectOtherData
//
#define PG_O_REPOS_START   1000
#define PG_O_REPOS        14000

#define PG_O_TOTAL_TEXT        "Loading miscellaneous data: Total Progress"
#define PG_O_REPOS_START_TEXT  "Reading repositories"
//      PG_O_REPOS_TEXT        "Downloading/Reading repository metadata for ..."

#define PG_O_TOTAL  (  PG_O_REPOS_START \
                     + PG_O_REPOS)

//
// Progress values for apply's Check Dependencies (i.e. before confirmation)
// No progress indications exist for this any longer, but continue to use the
// progress bar to show that it's doing something.
//
#define PG_D_RESOLVE  1

#define PG_D_TOTAL_TEXT        "Checking Dependencies: Total Progress"
#define PG_D_RESOLVE_TEXT      "Resolve Dependencies"
//      PG_D_DEPSOLVE_TEXT     "Downloading ... header"
//                             "Resolve Dependencies"


#define PG_D_TOTAL  (PG_D_RESOLVE)

//
// Progress values for apply's Run Transaction (i.e. after confirmation)
//
#define PG_R_INITIALIZE           100
#define PG_R_GRAB_RPMS_NOMINAL  24000 // used to define ts.run.pgRGrabRpms
#define PG_R_RUN_TRANS_TEST      1000
#define PG_R_RUN_TRANS           8000

#define PG_R_TOTAL_TEXT            "Running Transaction: Total Progress"
#define PG_R_INITIALIZE_TEXT       "Preparing"
//      PG_R_GRAB_RPMS             "Downloading ... header"
#define PG_R_RUN_TRANS_TEST_TEXT   "Preparing to Run Transaction"
//      PG_R_RUN_TRANS             "Transaction Test"
//                                 "Installing: ..."
//                                 "Updating: ..."
//                                 "Removing: ..."
//                                 "Completing update for: ..."

#define PG_R_TOTAL  (  PG_R_INITIALIZE     \
                     + ts.run.pgRGrabRpms  \
                     + PG_R_RUN_TRANS_TEST \
                     + PG_R_RUN_TRANS)

#define PG_R_GRAB(size)   (size)

// The number of acts for each operation based on the size of the package
//
// The size of an installed package is reported in uncompressed bytes.
// But the size of an uninstalled package is reported as its file size.
// Convert uninstalled package sizes by multiplying by 3 (a heuristic value).
// For updates, expect the removal portion to take a bit of extra time.
#define PG_R_INSTALL(size)          (max((size) * 3, PKG_MIN))
#define PG_R_REMOVE(size)           (max((size),     PKG_MIN))
#define PG_R_UPDATE(size)           (max((size) * 4, PKG_MIN))

class YumIO::EntryKey {
public:
   EntryKey(bool     isGroup,
            QString  name,
            QString  biarch)
      : isGroup(isGroup), 
        name(name),
        biarch(biarch)
   {}
   EntryKey(Package::Kind  kind,
            QString        name,
            QString        biarch)
      : isGroup(kind == Package::GROUP), 
        name(name),
        biarch(biarch)
   {}
   inline bool operator==(const EntryKey&  that) const
   {
      return (this->isGroup == that.isGroup
              && this->name == that.name
              && this->biarch == that.biarch);
   }
   bool     isGroup;
   QString  name;
   QString  biarch;
};

inline uint qHash(const YumIO::EntryKey&  key)
{
   return (qHash(key.name) << 1) + (qHash(key.biarch) << 2) + key.isGroup;
}

YumIO::YumIO()
   : clearing(false),
     logStartOp(-1),
     collectedFileLists(false),
     collectedOtherData(false),
     dialogs(NULL),
     capturers(NULL),
     rpm_dict(NULL),
     classYum(NULL),
     yumIO(NULL),
     base(NULL),
     reposDict(NULL),
     locked(false),
     yumVerKnown(false),
     totalProgressBar(NULL),
     itemProgressBar(NULL),
     pg(NULL)
{
   wr.stderrKind  = Log::ERROR;
   wr.stderrValue = 0;

   download.mode = Undefined;

   file_dbg = fdopen(dup(STDERR_FILENO), "w");
}

YumIO::~YumIO()
{
   fclose(file_dbg);
   file_dbg = NULL;

   Python::PyGILStateWrapper  gsw;

   if (initialized) {
      // Since the log window is toast now, and we're destroying the yumIO
      // now, which will render the YumIOStdout and YumIOStderr python
      // objects useless, let's change the python stdout & stderr to their
      // defaults.  That way, if any errors happen, they'll be seen and won't
      // cause either cascading hangs or SIGSEGV's.
      py("sys.stdout = sys.__stdout__\n"
         "sys.stderr = sys.__stderr__\n");
   }

   // To be proper, we'd call:
   //    Py_DECREF(yumIO);
   //    Py_DECREF(reposDict);
   // But this happens on exit, so don't waste the time.

   if (locked) {
      pyExceptStart();
      pyExceptExpect(xcpt.lockError);
      PyObject*  unlockStatus = pyMethodCall(yumIO, "unlock", "()");
      if (pyExceptMatches(unlockStatus, xcpt.lockError)) {
         pythonException(tr("Unable to release yum lock"));
      }
      Py_XDECREF(unlockStatus);
      pyExceptClear();
   }
}

void YumIO::setup(QWidget*       parent,
                  HProgressBar*  totalProgressBar,
                  HProgressBar*  itemProgressBar)
{
   dialogs = new YumIODialogs(parent);
   this->parent           = parent;
   this->totalProgressBar = totalProgressBar;
   this->itemProgressBar  = itemProgressBar;
}

bool YumIO::initialize()
{
   if (yumIO == NULL) {
      YumIOC::initialize(this);
   }
   Python::initialize();
   if (yumIO == NULL) {
      // Necessary initialization so that pyTime() can work.
      PyDateTime_IMPORT;

      Python::PyGILStateWrapper  gsw;

      // This is necessary for newer versions of yum (e.g. 2.4.2) when
      // calling getOptionsConfig in python.  And it's harmless for older
      // versions of yum (e.g. 2.2.2).
      PyObject*  yumArgv = Py_BuildValue("[s,s]", program, "info");
      pyAssign("sys", "argv", yumArgv);
      Py_DECREF(yumArgv);

      PyObject*  rpm_mod  = PyImport_ImportModule("rpm");
      if (rpm_mod == NULL) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: cannot find python rpm module"));
         return false;
      }

      rpm_dict = PyModule_GetDict(rpm_mod);
      if (rpm_dict == NULL) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: cannot find python rpm dict"));
         return false;
      }

      setEnablePythonSIGINTHandler(false);

      FILE*  yumIOFile
         = fopen(Share::absolute("python/YumIO.py").toStdString().c_str(), "r");
      if (yumIOFile == NULL) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: unable to open YumIO.py"));
         return false;
      }
      int  status = PyRun_SimpleFile(yumIOFile, "YumIO.py");
      if (status == -1) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: unable to execute YumIO.py"));
         return false;
      }
      fclose(yumIOFile);

      classYum = PyDict_GetItemString(main_dict, "YumIO");
      if (classYum == NULL) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: cannot find python YumIO"));
         return false;
      }
      pyExceptStart();
      pyExceptExpect(PyExc_SystemExit);
      PyObject*  args = Py_BuildValue("()");
      yumIO = PyObject_Call(classYum, args, NULL);
      Py_XDECREF(args);
      if (pyExceptMatches(yumIO, PyExc_SystemExit)) {
         pythonException(tr("Unable to initialize"));
         return false;
      } else if (yumIO == NULL) {
         if (PyErr_Occurred()) handleException();
         error(tr("Internal error: cannot create python yumIO"));
         return false;
      }

      base = PyObject_GetAttrString(yumIO, "base");
      if (base == NULL) {
         error(tr("Internal error: yumIO.base not evaluable"));
         return false;
      }

      xcpt.bdbQuit                 = pyLookup("bdb.BdbQuit");
      xcpt.yumBaseError            = pyLookup("dnf.exceptions.Error");
      xcpt.miscError               = pyLookup("dnf.exceptions.MiscError");
      xcpt.lockError               = pyLookup("dnf.exceptions.LockError");
      xcpt.groupsError             = NULL;
      xcpt.repoError               = pyLookup("dnf.exceptions.RepoError");
      xcpt.sqlite3OperationalError = NULL;
      xcpt.noMoreMirrorsRepoError  = NULL;
   }

   return true;
}

void  YumIO::blockerRelease ()
{
   blocker.release();
}

void  YumIO::execWorkThread (QEventLoop&  eventLoop,
                             QThread&     thread)
{
   blocker.prepare();

   PyThreadStateWrapper  tsw;

   thread.start(QThread::HighPriority);

   sigset_t  blocked;
   sigset_t  original;
   sigemptyset(&blocked);
   sigaddset(&blocked, SIGCHLD);
   sigprocmask(SIG_BLOCK, &blocked, &original);

   // Release the blocker once the eventLoop starts
   QTimer::singleShot(0, this, SLOT(blockerRelease()));

   eventLoop.exec();
   thread.wait();
   eventLoop.processEvents(QEventLoop::ExcludeUserInputEvents);

   // Undo the obnoxious SIGINT handler that either yum or rpm installed again
   {
      Python::PyGILStateWrapper  gsw;
      setEnablePythonSIGINTHandler(false);
   }

   sigprocmask(SIG_SETMASK, &original, NULL);
}

class CollectThread : public QThread 
{
public:
   CollectThread(YumIO*               yumIO,
                 QEventLoop&          eventLoop,
                 const YumReposUsed&  reposUsed)
      : yumIO(yumIO),
        eventLoop(eventLoop),
        reposUsed(reposUsed)
   {}
protected:
   YumIO*               yumIO;
   QEventLoop&          eventLoop;
   const YumReposUsed&  reposUsed;
public:
   bool                 result;
protected:
   void run()
   {
      result = yumIO->collectReal(reposUsed);
      eventLoop.exit();
   }
};

bool YumIO::collect (const YumReposUsed&  reposUsed)
{
   QEventLoop     eventLoop;
   CollectThread  thread(this, eventLoop, reposUsed);
   execWorkThread(eventLoop, thread);
   return thread.result;
}

bool YumIO::collectReal (const YumReposUsed&  reposUsed)
{
   blocker.block();

   Progress::User    pgUser(pg, totalProgressBar, itemProgressBar);
   YumIODialogsUser  yumIODialogsUser(dialogs, parent);

   log(Log::ACTION, 0, tr("Initialization"));

   Python::PyGILStateWrapper  gsw;

   // In dnf, file lists are collected automatically, always.
   // In dnf, it proved impossible to collect otherdata on demand, so
   // setupRepos ensures that it's collected during initialization.
   collectedFileLists = true;
   collectedOtherData = true;

   pg->startTotal(tr(PG_C_TOTAL_TEXT), PG_C_TOTAL);

   pg->startItem(tr(PG_C_INITIALIZE_TEXT), PG_C_INITIALIZE);
   if (!locked) {
      pyExceptStart();
      pyExceptExpect(xcpt.lockError);
      PyObject*  lockStatus = pyMethodCall(yumIO, "lock", "()");
      if (pyExceptMatches(lockStatus, xcpt.lockError)) {
         pythonException(tr("Unable to obtain yum lock"));
         pyExceptClear();
         pg->reset();
         return false;
      }
      Py_XDECREF(lockStatus);
      pyExceptClear();

      locked = true;
   }

   reposDict = PyObject_GetAttrString(base, "repos");
   if (reposDict == NULL) {
      error(tr("Internal error: yumIO.base.repos not evaluable"));
      return false;
   }
   if (!PyDict_Check(reposDict)) {
      error(tr("Internal error: yumIO.base.repos not a dictionary"));
      return false;
   }
   
   useRepos(base, reposUsed);

   Py_DECREF(base);

   download.mode = Collect;
   
   repoProgress.repos = collectRepos(reposDict);
   repoProgress.kind  = tr("contents");
   repoProgress.act   = 0;

   if (PyDict_Size(reposDict) == 0) {
      error(tr("No repositories are configured.\n"
               "Under the Repositories menu, Edit Configuration can be used"
               " to configure repositories."));
      download.mode = Undefined;
      return false;
   } else if (repoProgress.repos.count() == 0) {
      error(tr("No repositories are enabled.\n"
               "Under the Repositories menu, Edit Configuration can be used"
               " to configure repositories."));
      download.mode = Undefined;
      return false;
   }

   pg->startItem(tr(PG_C_DOWNLOAD_START_TEXT), PG_C_DOWNLOAD_START);

   pg->setupItemActs(PG_C_DOWNLOAD, repoProgress.repos.count());
   pyExceptStart();
   pyExceptExpect(xcpt.repoError);
   pyExceptExpect(xcpt.bdbQuit);
   PyObject*  reposStatus = pyMethodCall(yumIO, "setupRepos", "()");
   if (pyExceptMatches(reposStatus, xcpt.repoError)
       || pyExceptMatches(reposStatus, xcpt.bdbQuit)) {
      pythonExceptionV(tr("Unable to initialize"));
      pyExceptClear();
      pg->reset();
      download.mode = Undefined;
      return false;
   }
   Py_XDECREF(reposStatus);
   pyExceptClear();
   pg->endItemActs();

   basearch = getYumVar("basearch");

   pg->startItem(tr(PG_C_UPDATES_TEXT), PG_C_UPDATES);
   pyExceptStart();
   commExceptSetup();
   PyObject*  setupTsStatus = pyMethodCall(yumIO, "setupTs", "()");
   if (commExceptRaised(setupTsStatus, "setup transaction")) {
      pyExceptClear();
      pg->reset();
      download.mode = Undefined;
      return false;
   }
   Py_XDECREF(setupTsStatus);
   pyExceptClear();
   PyObject*  updates;
   {
#ifndef DEBUG_DISABLE_CAPTURERS
      YumIOCapturersUser  capturersUser(capturers, this);
#endif
      pyExceptStart();
      commExceptSetup();
      updates = pyMethodCall(yumIO, "collectUpdates", "()");
      if (commExceptRaised(updates, "collect updatable packages")) {
         pyExceptClear();
         pg->reset();
         download.mode = Undefined;
         return false;
      } else if (updates == NULL) {
         error(tr("Internal error: execution of yumIO.collectUpdates failed"));
         pyExceptClear();
         pg->reset();
         download.mode = Undefined;
         return false;
      }
      pyExceptClear();
   }

   if (!collectSet(hash, updates, Package::OUT_OF_DATE)) {
      download.mode = Undefined;
      return false;
   }

   Py_DECREF(updates);

   pg->startItem(tr(PG_C_INSTALL_TEXT), PG_C_INSTALL);
   PyObject*  info;
   {
#ifndef DEBUG_DISABLE_CAPTURERS
      YumIOCapturersUser  capturersUser(capturers, this);
#endif
      pyExceptStart();
      commExceptSetup();
      info = pyMethodCall(yumIO, "collectAll", "()");
      if (commExceptRaised(info, "collect installed and available packages")) {
         pyExceptClear();
         pg->reset();
         download.mode = Undefined;
         return false;
      } else if (info == NULL) {
         error(tr("Internal error: execution of yumIO.collectAll failed"));
         pyExceptClear();
         pg->reset();
         download.mode = Undefined;
         return false;
      }
      pyExceptClear();
   }

   if (   !collectField(hash, info, "installed", Package::INSTALLED)
       || !collectField(hash, info, "available", Package::NOT_INSTALLED)) {
      download.mode = Undefined;
      return false;
   }

   Py_DECREF(info);

   download.mode = Undefined;
      
   pg->startItem(tr(PG_C_GROUPS_TEXT), PG_C_GROUPS);
   pyExceptStart();
#if 0
   pyExceptExpect(xcpt.groupsError);
#endif
   commExceptSetup();
   PyObject*  setupTsGroupResult = pyMethodCall(yumIO, "setupTsGroup", "()");
#if 0
   if (pyExceptMatches(setupTsGroupResult, xcpt.groupsError)) {
      QString  value = pyStringDec(pyExceptValuef());
      if (value.contains("No Groups Available")) {
         // DO NOTHING: It's fairly normal.  (Although why it's lumped with 
         //             actually serious errors as the single 
         //             yum.Errors.GroupsErrors exception I have no idea!)
      } else {
         error(tr("Internal error:"
                  " execution of yumIO.setupTsGroup failed"));
         pyExceptClear();
         pg->reset();
         return false;
      }
   }
#endif
   if (commExceptRaised(setupTsGroupResult, "collect groups")) {
      pyExceptClear();
      pg->reset();
      return false;
   }
   Py_XDECREF(setupTsGroupResult);
   pyExceptClear();
   PyObject*  groupInfo = pyMethodCall(yumIO, "collectGroupInfo", "()");
   if (groupInfo == NULL) {
      error(tr("Internal error: execution of yumIO.collectGroupInfo failed"));
      return false;
   }

   if (!collectGroups(hash, groupInfo)) {
      return false;
   }

   PyObject*  collectDoneStatus = pyMethodCall(yumIO, "collectDone", "()");
   Py_DECREF(collectDoneStatus);

   Py_DECREF(groupInfo);

   EntryHash::const_iterator  it = hash.begin();
   for (; it != hash.end(); ++it) {
      list.append(*it);
   }

   return true;
}

void  YumIO::useRepos(PyObject*            repos,
                      const YumReposUsed&  reposUsed)
{
   YumReposUsed::const_iterator  it = reposUsed.begin();
   for (; it != reposUsed.end(); ++it) {
      PyObject*  repo = pyDictGetItem(reposDict, it.key());
      if (repo != NULL) {
         char*  id = strdup(it.key().toStdString().c_str());
         pyExceptStart();
         pyExceptExpect(xcpt.repoError);
         PyObject*  status;
         if (*it) {
            status = pyMethodCall(yumIO, "enableRepo", "(s)", id);
         } else {
            status = pyMethodCall(yumIO, "disableRepo", "(s)", id);
         }
         if (pyExceptMatches(status, xcpt.repoError)) {
            warning(tr("unable to setup repository %1:\n%2")
                       .arg(it.key()).arg(pyStringDec(pyExceptValuef())));
         }
         Py_XDECREF(status);
         pyExceptClear();
         free(id);
      }
   }
}

QStringList  YumIO::collectRepos (PyObject*  dict)
{
   QStringList  result;

   Py_ssize_t  pos = 0;
   for (;;) {
      PyObject*  key;
      PyObject*  value;
      if (!PyDict_Next(dict, &pos, &key, &value)) break;

      PyObject*  enabled = PyObject_GetAttrString(value, "enabled");
      if (enabled != NULL && PY_TO_LONG(enabled) != 0) {
         QString  name = pyString(key);
#if defined(DEBUG_COLLECT_DOWNLOAD) || defined(DEBUG_COLLECT_DOWNLOAD_COOKED)
         printf("collectRepos: %s\n", name.toStdString().c_str());
#endif
         result.append(name);
      }
      Py_XDECREF(enabled);
   }

   return result;
}

bool  YumIO::collectField (EntryHash&      hash,
                           PyObject*       info,
                           const char*     field,
                           Package::State  state)
{
   char*  copy = new char[strlen(field)+1];
   strcpy(copy, field);
   bool  result = collectField(hash, info, copy, state);
   delete [] copy;
   return result;
}

bool  YumIO::collectField (EntryHash&      hash,
                           PyObject*       info,
                           char*           field,
                           Package::State  state)
{
   PyObject*  set = PyObject_GetAttrString(info, field);
   if (set == NULL) {
      error(tr("Internal error: yumIO.collectAll.%1 not found").arg(field));
      return false;
   }

   bool  result = collectSet(hash, set, state);
   Py_DECREF(set);
   return result;
}

bool  YumIO::collectSet (EntryHash&      hash,
                         PyObject*       set,
                         Package::State  state)
{
   PyObject*  it = PyObject_GetIter(set);
   for (;;) {
      PyObject*  pkg = PyIter_Next(it);
      if (pkg == NULL) break;

      QString  name(pyStringDec(PyObject_GetAttrString(pkg, "name")));
      QString  arch(pyStringDec(PyObject_GetAttrString(pkg, "arch")));
      QString  biarch(getBiArch(arch));

      biarches.insert(biarch);

      EntryKey  key(false, name, biarch);
      if (!hash.contains(key)) {
         Package::Entry*&  entry = hash[key];
         bool  status = collectPkg(entry, state, pkg);
         if (!status) return false;
         if (namesBest.contains(name)) {
            Package::Entry*&  best = namesBest[name];
            PyObject*  result = pyMethodCall(yumIO, "isBetterThan", "(O,O)", 
                                             best->obj, entry->obj);
            if (result == Py_True) {
               entry->duplicateName = true;
            } else {
               best->duplicateName = true;
               best = entry;
            }
            Py_XDECREF(result);
         } else {
            namesBest[name] = entry;
         }
      } else {
         Package::Entry*  entry = hash[key];
         if (state == Package::INSTALLED) {
            // For an out-of-date package, in addition to the version available
            // for installation, keep track of the installed version.  For an
            // installed or out-of-date package, if there are multiple versions
            // installed, keep track of them all.  The Describe dialog and
            // tooltip displays everything in this list.
            while (entry->next != NULL) {
               entry = entry->next;
            }
            collectPkg(entry->next, state, pkg);
         }
      }

      Py_DECREF(pkg);
   }

   Py_DECREF(it);

   return true;
}

bool  YumIO::collectPkg (Package::Entry*&  entry,
                         Package::State    state,
                         PyObject*         pkg)
{
   entry = new Package::Entry;
   
   Py_INCREF(pkg);

   entry->action        = Package::NONE;
   entry->state         = state;
   entry->kind          = getPkgKind(pkg);
   entry->name          = pyStringDec(PyObject_GetAttrString(pkg, "name"));
   entry->verrel        = getVerRel(pkg);
   entry->repo          = simpleString(pkg, "repoid");
   entry->arch          = pyStringDec(PyObject_GetAttrString(pkg, "arch"));
   entry->biarch        = getBiArch(entry->arch);
   entry->size          = sizeOf(pkg);
   entry->sizeText      = bytesFormat(entry->size);
   entry->summary       = simpleString(pkg, "summary");
   entry->buildTime     = buildTimeOf(pkg);
   entry->buildTimeText = timeFormat(entry->buildTime);
   entry->obj           = pkg;


   return true;
}

bool  YumIO::collectGroups (EntryHash&  hash,
                            PyObject*   groupInfo)
{
   // Determine whether or not a group is installed by mimicking
   // GroupCommand._group_lists.installed_pred, which looks up the group id in
   // the self.base.history.group.
   PyObject*  history = PyObject_GetAttrString(base, "history");
   if (history == NULL) {
      error(tr("Internal error: yumIO.base.history not evaluable"));
      return false;
   }
   PyObject*  historyGroup = PyObject_GetAttrString(history, "group");
   if (historyGroup == NULL) {
      error(tr("Internal error: yumIO.base.history.group not evaluable"));
      return false;
   }

   PyObject*  list;

   if (hasDistribGroups()) {
      list = groupInfo;
   } else {
      list = PyObject_GetAttrString(groupInfo, "visible_groups");
      if (list == NULL) {
         error(tr("Internal error: groupInfo.visible_groups not evaluable"));
         return false;
      }
   }

   PyObject*  it = PyObject_GetIter(list);
   for (;;) {
      PyObject*  obj = PyIter_Next(it);
      if (obj == NULL) break;

      PyObject*  nameObj;
      PyObject*  groupObj;
      if (hasDistribGroups()) {
         PyObject*   user_visible = PyObject_GetAttrString(obj, "uservisible");
         if (user_visible == NULL) {
            PyErr_Clear();
            continue;
         }
         if (!PyObject_IsTrue(user_visible)) continue;
         nameObj = PyObject_GetAttrString(obj, "name");
         if (nameObj == NULL) {
            error(tr("Internal error: group without name"));
            return false;
         }
         groupObj = obj;
      } else {
         nameObj  = obj;
         groupObj = groupInfo;
      }

      QString   name = pyString(nameObj);
      EntryKey  key(true, name, QString());

      if (!hash.contains(key)) {
         bool  status = collectGroup(hash[key], groupObj, name, historyGroup);
         if (!status) return false;
      }

      if (hasDistribGroups()) {
         Py_DECREF(nameObj);
      }
      
      Py_DECREF(obj);
   }

   Py_DECREF(it);

   if (!hasDistribGroups()) {
      Py_DECREF(list);
   }

   Py_DECREF(historyGroup);
   Py_DECREF(history);

   return true;
}

bool  YumIO::collectGroup (Package::Entry*&  entry,
                           PyObject*         groupObj,
                           QString           name,
                           PyObject*         historyGroup)
{
   bool  installed;

   if (hasDistribGroups()) {
      char*  id = strdup(name.toStdString().c_str());
      PyObject*  installedObj = pyMethodCall(historyGroup, "get", "(s)", id);
      installed = (installedObj != Py_None);
      Py_XDECREF(installedObj);
      free(id);
   } else {
      // groupObj is a groupInfo
      PyObject*  groupInstalled = PyObject_GetAttrString(groupObj,
                                                         "group_installed");
      char*      nameCopy       = strdup(name.toStdString().c_str());
      PyObject*  installedObj   = PyDict_GetItemString(groupInstalled, nameCopy);
      free(nameCopy);
      if (installedObj == NULL) {
         error(tr("Internal error:"
                  " cannot find groupInfo.group_installed['%s']").arg(name));
         return false;
      }
      Py_DECREF(groupInstalled);
      installed = PyObject_IsTrue(installedObj);
   }

   Py_INCREF(groupObj);

   entry = new Package::Entry;
   
   entry->action = Package::NONE;
   entry->state  = (installed ? Package::INSTALLED 
                              : Package::NOT_INSTALLED);
   entry->kind   = Package::GROUP;
   entry->name   = name;
   entry->obj    = groupObj;

   if (entry->state == Package::INSTALLED && !upToDateGroup(*entry)) {
      entry->state = Package::OUT_OF_DATE;
   }

   return true;
}

void  YumIO::clear()
{
   {
      Python::PyGILStateWrapper  gsw;

      PyObject*  flushStatus;
      flushStatus = pyMethodCall(yumIO, "flush", "()");
      Py_XDECREF(flushStatus);
   }

   clearEntries();
}

class ClearEntriesThread : public QThread 
{
public:
   ClearEntriesThread(YumIO*       yumIO,
                      QEventLoop&  eventLoop)
      : yumIO(yumIO),
        eventLoop(eventLoop)
   {}
protected:
   YumIO*       yumIO;
   QEventLoop&  eventLoop;
protected:
   void run()
   {
      yumIO->clearEntriesReal();
      eventLoop.exit();
   }
};

void YumIO::clearEntries () 
{
   clearing = true;

   QEventLoop          eventLoop;
   ClearEntriesThread  thread(this, eventLoop);
   execWorkThread(eventLoop, thread);
}

void YumIO::clearEntriesReal ()
{
   blocker.block();

   Python::PyGILStateWrapper  gsw;

   {
      Package::EntryList::const_iterator  it = list.begin();
      for (; it != list.end(); ++it) {
         Package::Entry*  entry = *it;

         for (; entry != NULL; entry = entry->next) {
            Py_XDECREF(entry->obj);
            entry->obj = NULL;
         }
      }
   }

   qDeleteAll(list.begin(), list.end());
   list.clear();
   hash.clear();
   namesBest.clear();

   biarches.clear();

   clearing = false;
}

void YumIO::clearEntriesBeforeQuit () 
{
   // This leaks all the Package::Entry python objects, but it's no problem
   // since we're quitting anyway.

   clearing = true;

   Package::EntryList::const_iterator  it = list.begin();
   for (; it != list.end(); ++it) {
      Package::Entry*  entry = *it;

      for (; entry != NULL; entry = entry->next) {
         entry->obj = NULL;
      }
   }

   qDeleteAll(list.begin(), list.end());
   list.clear();
   hash.clear();
   namesBest.clear();

   biarches.clear();
}

QString  YumIO::getState (const Package::Entry&  entry)
{
   switch (entry.state) {
   case Package::OUT_OF_DATE:   return tr("Available for update");
   case Package::NOT_INSTALLED: return tr("Available for installation");
   case Package::INSTALLED:     return tr("Installed");
   default:                     return QString();
   }
}

QString  YumIO::getKind (const Package::Entry&  entry)
{
   switch (entry.kind) {
   case Package::SUITE:   return tr("Suite of packages");
   case Package::GROUP:   return tr("Group of packages");
   case Package::PACKAGE: return tr("Ordinary package");
   default:               return QString();
   }
}

QString  YumIO::getDescription (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "description");
}

QString  YumIO::getGroup (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "group");
}

QString  YumIO::getVendor (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "vendor");
}

QString  YumIO::getBuildHost (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "buildhost");
}

QString  YumIO::getInstallTime (const Package::Entry&  entry)
{
   return getRpmTime(entry.obj, "INSTALLTIME");
}

QString  YumIO::getBinaryRpm (const Package::Entry&  entry)
{
   // In newer yum versions, the relativepath field doesn't even exist.
   {
      Python::PyGILStateWrapper  gsw;
      if (!PyObject_HasAttrString(entry.obj, "relativepath")) return QString();
   }

   return simpleString(entry.obj, "relativepath");
}

QString  YumIO::getSourceRpm (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "sourcerpm");
}

QString  YumIO::getLicense (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "license");
}

QString  YumIO::getPackager (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "packager");
}

QString  YumIO::getUrl (const Package::Entry&  entry)
{
   return simpleString(entry.obj, "url");
}

QString  YumIO::getRelocations (const Package::Entry&  entry)
{
   return getRpmList(entry.obj, "PREFIXES");
}

QString  YumIO::getProvides (const Package::Entry&  entry)
{
   return getPrco(entry.obj, "PROVIDE", "provides");
}

QString  YumIO::getRequires (const Package::Entry&  entry)
{
   return getPrco(entry.obj, "REQUIRE", "requires");
}

QString  YumIO::getConflicts (const Package::Entry&  entry)
{
   return getPrco(entry.obj, "CONFLICT", "conflicts");
}

QString  YumIO::getObsoletes (const Package::Entry&  entry)
{
   return getPrco(entry.obj, "OBSOLETE", "obsoletes");
}

QString  YumIO::getFiles (const Package::Entry&  entry)
{
   Python::PyGILStateWrapper  gsw;

   QStringList  values;

   switch (entry.kind) {
   case Package::SUITE:
   case Package::PACKAGE:
      if (PyObject_HasAttrString(entry.obj, "files")) {
         // In current dnf-4.0.9.2, the files list always is available.  In
         // prior yum versions, that was not always so, so the collectFileLists
         // code would load it.  I'm retaining that code, even though it does
         // nothing currently, just because it might be needed again in the
         // future.
         bool  success;

         gsw.release();
         success = collectFileLists();
         gsw.ensure();

         if (success) {
            PyObject*  list = PyObject_GetAttrString(entry.obj, "files");
            if (list == NULL) {
               error(tr("Internal error: cannot find package's files"));
               return QString();
            }

            PyObject*  it = PyObject_GetIter(list);
            for (;;) {
               PyObject*  item = PyIter_Next(it);
               if (item == NULL) break;

               values.append(pyString(item));

               Py_DECREF(item);
            }

            Py_DECREF(it);
            Py_DECREF(list);
         }
      }
      break;
   case Package::GROUP:
      // DO NOTHING
      break;
   }

   return makeSortedUniq(values);
}

QString  YumIO::getChangeLog (const Package::Entry&  entry)
{
   Python::PyGILStateWrapper  gsw;

   QString  result;

   switch (entry.kind) {
   case Package::SUITE:
   case Package::PACKAGE:
      {
         // Must be a YumAvailablePackage.  First make sure we've collected
         // the File Lists, and then get the files from returnFileEntries()
         bool  success;

         gsw.release();
         success = collectOtherData();
         gsw.ensure();

         if (success) {
            PyObject*  list = PyObject_GetAttrString(entry.obj, "changelogs");
            if (list == NULL) {
               error(tr("Internal error: cannot find package's changelogs"));
               return QString();
            }
            if (list == Py_None) {
               Py_DECREF(list);
               return QString();
            }

            PyObject*  it = PyObject_GetIter(list);
            for (;;) {
               PyObject*  item = PyIter_Next(it);
               if (item == NULL) break;

               QString  time = pyDate(pyDictGetItem(item, "timestamp"));
               QString  name = pyString(pyDictGetItem(item, "author"));
               QString  text = pyString(pyDictGetItem(item, "text"));

               if (!result.isEmpty()) {
                  result.append("\n\n");
               }

               result.append("* ").append(time)
                     .append(' ').append(name)
                     .append('\n');
               result.append(text);

               Py_DECREF(item);
            }

            Py_DECREF(it);
            Py_DECREF(list);
         }
      }
      break;
   case Package::GROUP:
      // DO NOTHING
      break;
   }

   return result;
}

QString  YumIO::getRequiredGroups (const Package::Entry&  entry)
{
   return getGroupGrps(entry, "sub_groups");
}

QString  YumIO::getDefaultMetaPkgs (const Package::Entry&  entry)
{
   return getGroupGrps(entry, "default_metapkgs");
}

QString  YumIO::getOptionalMetaPkgs (const Package::Entry&  entry)
{
   return getGroupGrps(entry, "optional_metapkgs");
}

QString  YumIO::getMandatoryPkgs (const Package::Entry&  entry)
{
   return getGroupPkgs(entry, "mandatory_pkgs", "mandatory_packages");
}

QString  YumIO::getDefaultPkgs (const Package::Entry&  entry)
{
   return getGroupPkgs(entry, "default_pkgs", "default_packages");
}

QString  YumIO::getOptionalPkgs (const Package::Entry&  entry)
{
   return getGroupPkgs(entry, "optional_pkgs", "optional_packages");
}

QString  YumIO::getConditionalPkgs (const Package::Entry&  entry)
{
   if (hasDistribGroups()) {
      return getGroupPkgs(entry, "", "conditional_packages");
   } else {
      return QString();
   }
}

QString  YumIO::getYumVar (QString  name)
{
   PyObject*  conf   = PyObject_GetAttrString(base, "conf");
   if (conf == NULL) {
      if (PyErr_Occurred()) handleException();
      error(tr("Internal error: conf not evaluable"));
      return QString();
   }
   
   char*  nameStr = new char[name.length()+1];
   strcpy(nameStr, name.toStdString().c_str());

   PyObject*  resultObj = PyObject_GetAttrString(conf, nameStr);
   if (resultObj == NULL) {
      if (PyErr_Occurred()) handleException();
      error(tr("Internal error: yum var %1 not found").arg(name));
      return QString();
   }

   delete [] nameStr;

   Py_DECREF(conf);

   return pyStringDec(resultObj);
}

class CollectFileListsThread : public QThread 
{
public:
   CollectFileListsThread(YumIO*       yumIO,
                          QEventLoop&  eventLoop)
      : yumIO(yumIO),
        eventLoop(eventLoop)
   {}
protected:
   YumIO*       yumIO;
   QEventLoop&  eventLoop;
public:
   bool         result;
protected:
   void run()
   {
      result = yumIO->collectFileListsReal();
      eventLoop.exit();
   }
};

bool  YumIO::collectFileLists ()
{
   if (collectedFileLists) return true;

   QEventLoop               eventLoop;
   CollectFileListsThread   thread(this, eventLoop);
   execWorkThread(eventLoop, thread);
   return thread.result;
}

bool  YumIO::collectFileListsReal ()
{
   blocker.block();

   Progress::User    pgUser(pg, totalProgressBar, itemProgressBar);
   YumIODialogsUser  yumIODialogsUser(dialogs, parent);

   log(Log::ACTION, 0, tr("Collecting File Lists"));

   repoProgress.kind = tr("file lists");
   repoProgress.act  = 0;

   Python::PyGILStateWrapper  gsw;

   pg->startTotal(tr(PG_F_TOTAL_TEXT), PG_F_TOTAL);

   pg->startItem(tr(PG_F_REPOS_START_TEXT), PG_F_REPOS_START);

   pg->setupItemActs(PG_F_REPOS, repoProgress.repos.count());
   {
#ifndef DEBUG_DISABLE_CAPTURERS
      YumIOCapturersUser  capturersUser(capturers, this);
#endif
      pyExceptStart();
      commExceptSetup();
      PyObject*  info = pyMethodCall(yumIO, "collectFileLists", "()");
      if (commExceptRaised(info, "collect filelists")) {
         pyExceptClear();
         pg->reset();
         return false;
      } else if (info == NULL) {
         error(tr("Internal error:"
                  " execution of yumIO.collectFileLists failed"));
         pyExceptClear();
         pg->reset();
         return false;
      }
      Py_DECREF(info);
      pyExceptClear();
   }
   pg->endItemActs();

   collectedFileLists = true;

   emit collectedExtraData();

   return true;
}

class CollectOtherDataThread : public QThread 
{
public:
   CollectOtherDataThread(YumIO*       yumIO,
                          QEventLoop&  eventLoop)
      : yumIO(yumIO),
        eventLoop(eventLoop)
   {}
protected:
   YumIO*       yumIO;
   QEventLoop&  eventLoop;
public:
   bool         result;
protected:
   void run()
   {
      result = yumIO->collectOtherDataReal();
      eventLoop.exit();
   }
};

bool  YumIO::collectOtherData ()
{
   if (collectedOtherData) return true;

   QEventLoop               eventLoop;
   CollectOtherDataThread   thread(this, eventLoop);
   execWorkThread(eventLoop, thread);
   return thread.result;
}

bool  YumIO::collectOtherDataReal ()
{
   blocker.block();

   Progress::User    pgUser(pg, totalProgressBar, itemProgressBar);
   YumIODialogsUser  yumIODialogsUser(dialogs, parent);

   log(Log::ACTION, 0, tr("Collecting Miscellaneous data"));

   repoProgress.kind = tr("miscellaneous data");
   repoProgress.act  = 0;

   Python::PyGILStateWrapper  gsw;

   pg->startTotal(tr(PG_O_TOTAL_TEXT), PG_O_TOTAL);

   pg->startItem(tr(PG_O_REPOS_START_TEXT), PG_O_REPOS_START);

   pg->setupItemActs(PG_O_REPOS, repoProgress.repos.count());
   {
#ifndef DEBUG_DISABLE_CAPTURERS
      YumIOCapturersUser  capturersUser(capturers, this);
#endif
      pyExceptStart();
      commExceptSetup();
      PyObject*  info = pyMethodCall(yumIO, "collectOtherData", "()");
      if (commExceptRaised(info, "collect miscellaneous data")) {
         pyExceptClear();
         pg->reset();
         return false;
      } else if (info == NULL) {
         error(tr("Internal error:"
                  " execution of yumIO.collectOtherData failed"));
         pyExceptClear();
         pg->reset();
         return false;
      }
      Py_DECREF(info);
      pyExceptClear();
   }
   pg->endItemActs();

   collectedOtherData = true;

   emit collectedExtraData();

   return true;
}

bool  YumIO::upToDateGroup (const Package::Entry&  entry)
{
   return (upToDateGroupGrps(entry, "sub_groups")
           && upToDateGroupPkgs(entry, "mandatory_pkgs", "mandatory_packages"));
}

bool  YumIO::upToDateGroupGrps (const Package::Entry&  entry,
                                const char*            attr_22)
{
   char*  copy = new char[strlen(attr_22)+1];
   strcpy(copy, attr_22);
   bool  result = upToDateGroupGrps(entry, copy);
   delete [] copy;
   return result;
}

bool  YumIO::upToDateGroupGrps (const Package::Entry&  entry,
                                char*                  attr_22)
{
   if (hasDistribGroups()) {
      // DO NOTHING
   } else {
      PyObject*  seq = getGroupItem_22(entry, attr_22);
      if (seq == NULL) {
         return true;
      }
      if (PySequence_Length(seq) > 0) {
         QStringList  values;

         PyObject*  it = PyObject_GetIter(seq);
         for (;;) {
            PyObject*  grpId = PyIter_Next(it);
            if (grpId == NULL) break;
            EntryKey  key(true, pyString(grpId), QString());
            if (hash.contains(key)) {
               if (!upToDateGroup(*hash[key])) {
                  Py_DECREF(seq);
                  return false;
               }
            }
            Py_DECREF(grpId);
         }

         Py_DECREF(it);
      }
      Py_DECREF(seq);
   }

   return true;
}

bool  YumIO::upToDateGroupPkgs (const Package::Entry&  entry,
                                const char*            attr_22,
                                const char*            attr_25)
{
   char*  attr_22_copy = new char[strlen(attr_22)+1];
   char*  attr_25_copy = new char[strlen(attr_25)+1];
   strcpy(attr_22_copy, attr_22);
   strcpy(attr_25_copy, attr_25);
   bool  result = upToDateGroupPkgs(entry, attr_22_copy, attr_25_copy);
   delete [] attr_25_copy;
   delete [] attr_22_copy;
   return result;
}

bool  YumIO::upToDateGroupPkgs (const Package::Entry&  entry,
                                char*                  attr_22,
                                char*                  attr_25)
{
   if (hasDistribGroups()) {
      PyObject*  dict = getGroupItem_25(entry, attr_25);
      if (dict == NULL) {
         return true;
      }

      if (PyDict_Size(dict) > 0) {
         Py_ssize_t  pos = 0;
         for (;;) {
            PyObject*  pkgName;
            PyObject*  value;
            if (!PyDict_Next(dict, &pos, &pkgName, &value)) break;

            Package::Entry*  pkg = findPackage(pyString(pkgName));
            if (pkg != NULL) {
               if (pkg->state == Package::OUT_OF_DATE) {
                  Py_DECREF(dict);
                  return false;
               }
            }
         }
      }

      Py_DECREF(dict);
   } else {
      PyObject*  seq = getGroupItem_22(entry, attr_22);
      if (seq == NULL) {
         return true;
      }

      if (PySequence_Length(seq) > 0) {
         PyObject*  it = PyObject_GetIter(seq);
         for (;;) {
            PyObject*  pkgName = PyIter_Next(it);
            if (pkgName == NULL) break;
            Package::Entry*  pkg = findPackage(pyString(pkgName));
            if (pkg != NULL) {
               if (pkg->state == Package::OUT_OF_DATE) {
                  Py_DECREF(seq);
                  return false;
               }
            }
            Py_DECREF(pkgName);
         }

         Py_DECREF(it);
      }
      Py_DECREF(seq);
   }

   return true;
}

const char*  YumIO::decodePrcoFlags (PyObject*  flags)
{
   char*        flagsCode = Python::pyString(flags);
   const char*  result    = YUM_DECODE_PRCO_FLAGS(flagsCode);
   free(flags);
   return result;
}

QString  YumIO::getPrco (PyObject*    pkg,
                         const char*  hdrKey,
                         const char*  prcoKey)
{
   char*  procKeyCopy = new char[strlen(prcoKey)+1];
   strcpy(procKeyCopy, prcoKey);
   QString  result = getPrco(pkg, hdrKey, procKeyCopy);
   delete [] procKeyCopy;
   return result;
}

QString  YumIO::getPrco (PyObject*    pkg,
                         const char*  hdrKey,
                         char*        prcoKey)
{
   Python::PyGILStateWrapper  gsw;

   QStringList  values;

   if (PyObject_HasAttrString(pkg, "hdr")) {
      // Must be a YumInstalledPackage.  Get the prco from the hdr
      char  nameKey[1024];
      char  flagsKey[1024];
      char  versionKey[1024];
      sprintf(nameKey,    "%sNAME",    hdrKey);
      sprintf(flagsKey,   "%sFLAGS",   hdrKey);
      sprintf(versionKey, "%sVERSION", hdrKey);

      PyObject*  nameList    = getRpmHdrValue(pkg, nameKey);
      if (nameList == NULL || nameList == Py_None) return QString();
      PyObject*  flagsList   = getRpmHdrValue(pkg, flagsKey);
      PyObject*  versionList = getRpmHdrValue(pkg, versionKey);

      PyObject*  nameIt    = PyObject_GetIter(nameList);
      PyObject*  flagsIt   = (flagsList != NULL ? PyObject_GetIter(flagsList)
                                                : NULL);
      PyObject*  versionIt = (flagsList != NULL ? PyObject_GetIter(versionList)
                                                : NULL);
      for (;;) {
         PyObject*  nameItem    = PyIter_Next(nameIt);
         if (nameItem == NULL) break;
         PyObject*  flagsItem   = (flagsIt != NULL   ? PyIter_Next(flagsIt)
                                                     : NULL);
         PyObject*  versionItem = (versionIt != NULL ? PyIter_Next(versionIt)
                                                     : NULL);

         QString  value = pyString(nameItem);

         if (   flagsItem   != NULL && flagsItem   != Py_None
             && versionItem != NULL && versionItem != Py_None) {
            unsigned long  flags = PY_TO_LONG(flagsItem);
            QString        flagsStr;
            if (flags & RPMSENSE_LESS)    flagsStr.append('<');
            if (flags & RPMSENSE_GREATER) flagsStr.append('>');
            if (flags & RPMSENSE_EQUAL)   flagsStr.append('=');
            value.append(' ').append(flagsStr);
            value.append(' ').append(pyString(versionItem));
         }

         values.append(value);

         Py_XDECREF(versionItem);
         Py_XDECREF(flagsItem);
         Py_DECREF(nameItem);
      }      

      Py_XDECREF(versionIt);
      Py_XDECREF(flagsIt);
      Py_DECREF(nameIt);

      Py_DECREF(versionList);
      Py_DECREF(flagsList);
      Py_DECREF(nameList);
   } else if (PyObject_HasAttrString(pkg, "prco")) {
      // Must be a YumAvailablePackage.  
      //
      // If the package has a returnPrco function, as is the case with newer
      // versions of yum (e.g. 2.4.2), then call it.  This is necessary even
      // if they have a prco attribute, because it is filled in lazily.
      //
      // However, for older versions of yum (e.g. 2.2.2), there is no
      // returnPrco function, so use the prco attribute directly.  Note that
      // returnSimple won't return the prco attribute.
      PyObject*  list;
      if (PyObject_HasAttrString(pkg, "returnPrco")) {
         list = pyMethodCall(pkg, "returnPrco", "(s)", prcoKey);
      } else {
         PyObject*  prco = PyObject_GetAttrString(pkg, "prco");
         list = PyDict_GetItemString(prco, prcoKey);
         Py_DECREF(prco);
      }
      if (list == NULL) {
         error(tr("Internal error:"
                  " cannot find package's prco.%1").arg(prcoKey));
         return QString();
      }

      PyObject*  it = PyObject_GetIter(list);
      for (;;) {
         PyObject*  item = PyIter_Next(it);
         if (item == NULL) break;

         PyObject*  name  = PyTuple_GetItem(item, 0);
         PyObject*  flags = PyTuple_GetItem(item, 1);
         PyObject*  evr   = PyTuple_GetItem(item, 2);

         QString  value = pyString(name);
         
         if (   flags != NULL && flags != Py_None
             && evr   != NULL && evr   != Py_None) {
            const char*  flagsStr = decodePrcoFlags(flags);
            PyObject*    epoch    = PyTuple_GetItem(evr, 0);
            PyObject*    ver      = PyTuple_GetItem(evr, 1);
            PyObject*    rel      = PyTuple_GetItem(evr, 2);
            if (ver != NULL && ver != Py_None) {
               value.append(' ').append(flagsStr).append(' ');
               if (epoch != NULL && epoch != Py_None) {
                  char*  epochStr = Python::pyString(epoch);
                  if (strcmp(epochStr, "0") != 0) {
                     value.append(epochStr).append(':');
                  }
                  free(epochStr);
               }
               value.append(pyString(ver));
               if (rel != NULL && rel != Py_None) {
                  value.append('-').append(pyString(rel));
               }
            }
         }

         values.append(value);

         Py_DECREF(item);
      }

      Py_DECREF(it);
      if (PyObject_HasAttrString(pkg, "returnPrco")) {
         Py_DECREF(list);
      }
   }

   return makeSorted(values);
}

#if 0
QString  YumIO::getRpm(PyObject*  pkg,
                       char*      rpmKey)
{
   Python::PyGILStateWrapper  gsw;

   if (PyObject_HasAttrString(pkg, "hdr")) {
      return pyStringXDec(getRpmHdrValue(pkg, rpmKey));
   } else {
      return QString();
   }
}
#endif

QString  YumIO::getRpmTime(PyObject*    pkg,
                           const char*  rpmKey)
{
   Python::PyGILStateWrapper  gsw;

   if (PyObject_HasAttrString(pkg, "hdr")) {
      return timeFormat(pyTimeDec(getRpmHdrValue(pkg, rpmKey)));
   } else {
      return QString();
   }
}

QString  YumIO::getRpmList(PyObject*    pkg,
                           const char*  rpmKey)
{
   Python::PyGILStateWrapper  gsw;

   if (PyObject_HasAttrString(pkg, "hdr")) {
      QString  value;

      PyObject*  list = getRpmHdrValue(pkg, rpmKey);
      if (list != NULL && list != Py_None) {
         PyObject*  it = PyObject_GetIter(list);
         for (;;) {
            PyObject*  item = PyIter_Next(it);
            if (item == NULL) break;
            if (!value.isEmpty()) {
               value.append('\n');
            }
            value.append(pyString(item));
            Py_DECREF(item);
         }
         Py_DECREF(it);
         Py_DECREF(list);
      }

      return value;
   } else {
      return QString();
   }
}

QString  YumIO::getGroupGrps (const Package::Entry&  entry,
                              const char*            attr_22)
{
   char*  copy = new char[strlen(attr_22)+1];
   strcpy(copy, attr_22);
   QString  result = getGroupGrps(entry, copy);
   delete [] copy;
   return result;
}

QString  YumIO::getGroupGrps (const Package::Entry&  entry,
                              char*                  attr_22)
{
   Python::PyGILStateWrapper  gsw;

   if (hasDistribGroups()) {
      return QString();
   } else {
      PyObject*  seq = getGroupItem_22(entry, attr_22);
      if (seq == NULL) {
         return QString();
      }

      if (PySequence_Length(seq) > 0) {
         QStringList  values;

         PyObject*  it = PyObject_GetIter(seq);
         for (;;) {
            PyObject*  grpId = PyIter_Next(it);
            if (grpId == NULL) break;
            values.append(pyString(grpId));
            Py_DECREF(grpId);
         }

         Py_DECREF(it);

         Py_DECREF(seq);
         return makeSorted(values);
      } else {
         Py_DECREF(seq);
         return QString();
      }
   }
}

QString  YumIO::getGroupPkgs (const Package::Entry&  entry,
                              const char*            attr_22,
                              const char*            attr_25)
{
   char*  attr_22_copy = new char[strlen(attr_22)+1];
   char*  attr_25_copy = new char[strlen(attr_25)+1];
   strcpy(attr_22_copy, attr_22);
   strcpy(attr_25_copy, attr_25);
   QString  result = getGroupPkgs(entry, attr_22_copy, attr_25_copy);
   delete [] attr_25_copy;
   delete [] attr_22_copy;
   return result;
}

QString  YumIO::getGroupPkgs (const Package::Entry&  entry,
                              char*                  attr_22,
                              char*                  attr_25)
{
   Python::PyGILStateWrapper  gsw;

   if (hasDistribGroups()) {
      PyObject*  dict = getGroupItem_25(entry, attr_25);
      if (dict == NULL) {
         return QString();
      }

      if (PyDict_Size(dict) > 0) {
         QStringList  values;

         Py_ssize_t  pos = 0;
         for (;;) {
            PyObject*  pkgName;
            PyObject*  value;
            if (!PyDict_Next(dict, &pos, &pkgName, &value)) break;

            values.append(pyString(pkgName));
         }

         Py_DECREF(dict);
         return makeSorted(values);
      } else {
         Py_DECREF(dict);
         return QString();
      }
   } else {
      PyObject*  seq = getGroupItem_22(entry, attr_22);
      if (seq == NULL) {
         return QString();
      }

      if (PySequence_Length(seq) > 0) {
         QStringList  values;

         PyObject*  it = PyObject_GetIter(seq);
         for (;;) {
            PyObject*  pkgName = PyIter_Next(it);
            if (pkgName == NULL) break;
            values.append(pyString(pkgName));
            Py_DECREF(pkgName);
         }

         Py_DECREF(it);

         Py_DECREF(seq);
         return makeSorted(values);
      } else {
         Py_DECREF(seq);
         return QString();
      }
   }
}

QString  YumIO::makeSorted (QStringList&  list)
{
   qSort(list.begin(), list.end());

   QString  result;

   QStringList::const_iterator  it = list.begin();
   if (it != list.end()) {
      result.append(*it); ++it;
   }
   for (; it != list.end(); ++it) {
      result.append("\n").append(*it);
   }

   return result;
}

QString  YumIO::makeSortedUniq (QStringList&  list)
{
   qSort(list.begin(), list.end());

   QString  result;

   QStringList::const_iterator  it   = list.begin();
   QStringList::const_iterator  prev = list.begin();
   if (it != list.end()) {
      result.append(*it); prev = it; ++it;
   }
   for (; it != list.end(); prev = it, ++it) {
      if (*it != *prev) {
         result.append("\n").append(*it);
      }
   }

   return result;
}

PyObject*  YumIO::getRpmHdrValue(PyObject*    pkg,
                                 const char*  rpmKey)
{
   if (!PyObject_HasAttrString(pkg, "hdr")) {
      error(tr("getRpmHdrValue called for package with no hdr"
               " (i.e. not installed)"));
      return NULL;
   }

   PyObject*  hdr = PyObject_GetAttrString(pkg, "hdr");

   char  key[1024];
   sprintf(key, "RPMTAG_%s", rpmKey);

   PyObject*  tag = PyDict_GetItemString(rpm_dict, key);
   if (tag == NULL) {
      error(tr("Internal error: cannot find rpm.%1").arg(key));
      return NULL;
   }

   PyObject*  result = PyObject_GetItem(hdr, tag);
   Py_DECREF(hdr);
   return result;
}

class ApplyThread : public QThread 
{
public:
   ApplyThread(YumIO*       yumIO,
               QEventLoop&  eventLoop)
      : yumIO(yumIO),
        eventLoop(eventLoop)
   {}
protected:
   YumIO*       yumIO;
   QEventLoop&  eventLoop;
public:
   bool         result;
protected:
   void run()
   {
      result = yumIO->applyReal();
      eventLoop.exit();
   }
};

bool YumIO::apply () 
{
   QEventLoop   eventLoop;
   ApplyThread  thread(this, eventLoop);
   execWorkThread(eventLoop, thread);
   return thread.result;
}

bool YumIO::applyReal () 
{
   blocker.block();

   Progress::User    pgUser(pg, totalProgressBar, itemProgressBar);
   YumIODialogsUser  yumIODialogsUser(dialogs, parent);

   log(Log::ACTION, 0, tr("Applying"));

   if (geteuid() != 0) {
      error("Must be run as root to apply changes");
      return false;
   }

   Python::PyGILStateWrapper  gsw;

   pg->startTotal(tr(PG_D_TOTAL_TEXT), PG_D_TOTAL);

   pg->startItem(tr(PG_D_RESOLVE_TEXT), PG_D_RESOLVE);

   download.mode = Ts;
   
#if defined(DEBUG_TS_DESCR)
   dprintf("ts.descr: -----clear-----\n");
#endif

   ts.descr               = QString();
   ts.list.during         = false;
   ts.list.confirmed      = false;
   ts.act                 = 0;
   ts.run.doingTest       = false;
   ts.run.nominalGrabSize = 0;
   ts.run.totalGrabSize   = 0;
   ts.run.totalTsSize     = 0;
   ts.run.lastTs          = 0;
   ts.run.lastAction      = -1;
   ts.run.deltaRebuilt    = 0;
   ts.parallel.begun      = false;
   ts.parallel.meterMapping.clear();

   PyObject*  installs = PyList_New(0);
   PyObject*  updates  = PyList_New(0);
   PyObject*  removes  = PyList_New(0);

   {
      Package::EntryList::const_iterator  it = list.begin();
      for (; it != list.end(); ++it) {
         Package::Entry*  entry = *it;
         entry->apply.action = entry->action;
         entry->apply.reason = entry;
      }
   }

   bool  status = applyGroups();
   if (!status) {
      download.mode = Undefined;
      return false;
   }

   // I'm not bothering to Py_INCREF and Py_DECREF the packages themselves,
   // because I already am holding a reference for them because they're in
   // the entries list.  That's good enough.

   {
      Package::EntryList::const_iterator  it = list.begin();
      for (; it != list.end(); ++it) {
         Package::Entry*  entry = *it;

         if (entry->kind != Package::GROUP) {
            switch (entry->apply.action) {
            case Package::UPDATE:
               {
                  if (entry->next == NULL) {
                     error(tr("Unable to find installed version for package"
                              " update %1").arg(entry->name));
                     download.mode = Undefined;
                     return false;
                  }
                  // The first entry in the list (entry) is the new package.
                  // All subsequent entries (e) are old packages to be removed.
                  // Usually, there is only one of those, but it's possible for
                  // there to be more.
                  Package::Entry*  e = entry->next;
                  for (; e != NULL; e = e->next) {
                     PyObject*  pair = PyList_New(0);
                     status = PyList_Append(pair, entry->obj);
                     if (status != 0) {
                        error(tr("Unable to add new package to pair for updated"
                                 " package %1").arg(entry->name));
                     }
                     status = PyList_Append(pair, e->obj);
                     if (status != 0) {
                        error(tr("Unable to add old package to pair for updated"
                                 " package %1").arg(entry->name));
                     }
                     status = PyList_Append(updates, pair);
                     if (status != 0) {
                        error(tr("Unable to add package %1 to updates"
                                 " list").arg(entry->name));
                     }
                  }
               }
               break;
            case Package::INSTALL:
               {
                  int  status = PyList_Append(installs, entry->obj);
                  if (status != 0) {
                     error(tr("Unable to add package %1 to installs"
                              " list").arg(entry->name));
                  }
               }
               break;
            case Package::REMOVE:
               {
                  Package::Entry*  e = entry;
                  if (e->state == Package::OUT_OF_DATE) {
                     // Skip the updateable package, since it isn't actually
                     // installed currently.
                     e = e->next;
                  }
                  for (; e != NULL; e = e->next) {
                     int  status = PyList_Append(removes, e->obj);
                     if (status != 0) {
                        error(tr("Unable to add package %1 to removes"
                                 " list").arg(entry->name));
                     }
                  }
               }
               break;
            case Package::NONE:
               // DO NOTHING
               break;
            }
         }
      }
   }

   bool  success = reapChildren();
   if (!success) {
      pg->reset();
      download.mode = Undefined;
      return false;
   }

   {
#ifndef DEBUG_DISABLE_CAPTURERS
      YumIOCapturersUser  capturersUser(capturers, this);
#endif
      sigchldDisable();
      pyExceptStart();
      commExceptSetup();
      PyObject*  applyStatus = pyMethodCall(yumIO, "apply", "(O,O,O)", 
                                            installs, updates, removes);
      if (commExceptRaised(applyStatus, "apply changes")) {
         pyExceptClear();

         // Set ts.list.confirmed so we'll return false, and let the rest of
         // this routine do its clean up.
         ts.list.confirmed = false;

         // At least one major purpose of this is to call self.closeRpmDB ->
         // self._ts.close -> self._ts.ts.closeDB -> rpmts_CloseDB -> 
         // rpmdbClose -> rpmsqEnable, which prevents rpmdbCheckSignals
         // from calling exit(1) if a SIGINT caused apply to fail.
         PyObject*  applyFailCleanupStatus;
         applyFailCleanupStatus = pyMethodCall(yumIO, "applyFailCleanup", "()");
         Py_XDECREF(applyFailCleanupStatus);
      } else {
         Py_XDECREF(applyStatus);
         pyExceptClear();
      }

      sigchldRestore();
   }

   download.mode = Undefined;

   PyObject*  flushStatus;
   flushStatus = pyMethodCall(yumIO, "flush", "()");
   Py_XDECREF(flushStatus);

   Py_DECREF(removes);
   Py_DECREF(updates);
   Py_DECREF(installs);

   pg->reset();

   return ts.list.confirmed;
}

bool  YumIO::cleanHeaders ()
{
   log(Log::ACTION, 0, tr("Cleaning Headers"));

   Python::PyGILStateWrapper  gsw;

   PyObject*  pair = pyMethodCall(yumIO, "cleanHeaders", "()");

   bool  result = (PY_TO_LONG(PyTuple_GetItem(pair, 0)) == 0);
   Py_XDECREF(pair);
   return result;
}

bool  YumIO::cleanPackages ()
{
   log(Log::ACTION, 0, tr("Cleaning Packages"));

   Python::PyGILStateWrapper  gsw;

   PyObject*  pair = pyMethodCall(yumIO, "cleanPackages", "()");

   bool  result = (PY_TO_LONG(PyTuple_GetItem(pair, 0)) == 0);
   Py_XDECREF(pair);
   return result;
}

bool  YumIO::cleanXmlMetadata ()
{
   log(Log::ACTION, 0, tr("Cleaning Xml Metadata"));

   Python::PyGILStateWrapper  gsw;

   PyObject*  pair = pyMethodCall(yumIO, "cleanMetadata", "()");

   bool  result = (PY_TO_LONG(PyTuple_GetItem(pair, 0)) == 0);
   Py_XDECREF(pair);
   return result;
}

bool  YumIO::cleanPickledCache ()
{
   log(Log::ACTION, 0, tr("Cleaning Pickled Cache"));

   Python::PyGILStateWrapper  gsw;

   PyObject*  pair = pyMethodCall(yumIO, "cleanPickles", "()");

   bool  result = (PY_TO_LONG(PyTuple_GetItem(pair, 0)) == 0);
   Py_XDECREF(pair);
   return result;
}

bool  YumIO::cleanSqliteCache ()
{
   log(Log::ACTION, 0, tr("Cleaning Sqlite Cache"));

   Python::PyGILStateWrapper  gsw;

   PyObject*  pair = pyMethodCall(yumIO, "cleanSqlite", "()");

   bool  result = (PY_TO_LONG(PyTuple_GetItem(pair, 0)) == 0);
   Py_XDECREF(pair);
   return result;
}

bool  YumIO::cleanAll ()
{
   log(Log::ACTION, 0, tr("Cleaning All"));

   Python::PyGILStateWrapper  gsw;

   PyObject*  hdrPair  = pyMethodCall(yumIO, "cleanHeaders", "()");
   PyObject*  pkgPair  = pyMethodCall(yumIO, "cleanPackages", "()");
   PyObject*  metaPair = pyMethodCall(yumIO, "cleanMetadata", "()");

   long  hdrValue  = PY_TO_LONG(PyTuple_GetItem(hdrPair, 0));
   long  pkgValue  = PY_TO_LONG(PyTuple_GetItem(pkgPair, 0));
   long  metaValue = PY_TO_LONG(PyTuple_GetItem(metaPair, 0));
   long  pickValue = 0;
   long  sqlValue  = 0;
   
   Py_XDECREF(hdrPair);
   Py_XDECREF(pkgPair);
   Py_XDECREF(metaPair);

   {
      PyObject*  sqlPair = pyMethodCall(yumIO, "cleanSqlite", "()");
      sqlValue = PY_TO_LONG(PyTuple_GetItem(sqlPair, 0));
      Py_XDECREF(sqlPair);
   }

   return (hdrValue == 0
           && pkgValue == 0
           && metaValue == 0
           && pickValue == 0
           && sqlValue == 0);
}

//
// stdoutWrite is for output intercepted by YumIOStdout.write.
// stderrWrite is for output intercepted by YumIOStderr.write.
//
// These are the best methods of I/O interception because they are synchronous.
// But if they fail, the YumIOCapturers mechanism will intercept the rest.
//

PyObject*  YumIO::stdoutWrite(PyObject*  self, PyObject*  args)
{ 
   return stdWrite(self, args, Log::STDOUT, 0, wr.stdoutBuffer);
}

PyObject*  YumIO::stderrWrite(PyObject*  self, PyObject*  args)
{ 
   return stdWrite(self, args, wr.stderrKind, wr.stderrValue, wr.stderrBuffer);
}

PyObject*  YumIO::stdWrite(PyObject*     self, 
                           PyObject*     args,
                           Log::Kind     kind,
                           unsigned int  value,
                           QString&      buffer)
{
   char*  content;
   
   if (!PyArg_ParseTuple(args, "s", &content)) return NULL;

#if defined(DEBUG_IO)
   dprintf("stdWrite: %s\n", content);
#endif

   if (ts.list.during) {
      // While during the package list, all text for the query is NUU-generated.
   } else {
      // Accumulate text just in case it is part of a yum query.
#if defined(DEBUG_TS_DESCR)
      dprintf("ts.descr: %s\n", content);
#endif
      ts.descr.append(content).append("\n");
   }

   write(content, kind, value, buffer);

   Py_INCREF(Py_None);
   return Py_None;
}

void  YumIO::write (QString       content,
                    Log::Kind     kind,
                    unsigned int  value,
                    QString&      buffer)
{
   buffer.append(content);

   QStringList  lines = buffer.split("\n");

   QStringList::const_iterator  it   = lines.begin();
   QStringList::const_iterator  last = lines.end() - 1;

   if (it != last) {
      logEntries.lock();

      int i = 0;
      for (; it != last; ++it, ++i) {
         logEntries.append(Log::Entry(kind, value, *it));
      }

      logEntries.unlock();
      logEntries.emitAppended();
   }

   // The last line in the list is, by definition, unfinished.  (Note that
   // a very common case would be when it's the empty string.)  Sit on it
   // until further text comes along to finish it with a newline.
   buffer = *it;
}

//
// stdoutCapturedWrite and stderrCapturedWrite are for output which bypassed the
// YumIOStdout/YumIOStderr mechanism and had to be intercepted by
// YumIOCapturersMinion.
//
// These are the ultimate methods of I/O interception, but they do suffer from
// delays based on buffered I/O.  The I/O here typically would be produced by
// Python (although not necessarily), then sitting in a C FILE* buffer for a
// while, then sitting in the YumIOCapturers pipe buffer for a while, before
// finally arriving here.
//

void  YumIO::stdoutCapturedWrite (QString  content)
{
#if defined(DEBUG_IO)
   dprintf("stdoutCapturedWrite: %s\n", content.toStdString().c_str());
#endif
   write(content, Log::STDOUT, 0, wr.stdoutBuffer);
}

void  YumIO::stderrCapturedWrite (QString  content)
{
#if defined(DEBUG_IO)
   dprintf("stderrCapturedWrite: %s\n", content.toStdString().c_str());
#endif
   write(content, Log::ERROR, 0, wr.stderrBuffer);
}

PyObject*  YumIO::log(PyObject*  self, PyObject*  args)
{ 
   long   value;
   char*  msg;
   
   if (!PyArg_ParseTuple(args, "ls", &value, &msg)) return NULL;

   log(Log::NORMAL, value, msg);

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::errorlog(PyObject*  self, PyObject*  args)
{ 
   long   value;
   char*  msg;
   
   if (!PyArg_ParseTuple(args, "ls", &value, &msg)) return NULL;

   log(Log::ERROR, value, msg);

   Py_INCREF(Py_None);
   return Py_None;
}

void  YumIO::log (Log::Kind     kind,
                  unsigned int  value,
                  QString       text)
{
#if defined(DEBUG_IO)
   dprintf("log: %s\n", text.toStdString().c_str());
#endif

   QString  edited(text);
   edited.replace(QRegExp(": (\\[Errno [0-9]*\\] )"), ":\n\\1");

   QStringList  lines = edited.split("\n");
   
   logEntries.lock();

   QStringList::const_iterator  it = lines.begin();
   for (; it != lines.end(); ++it) {
      logEntries.append(Log::Entry(kind, value, *it));
   }
   
   logEntries.unlock();
   logEntries.emitAppended();
}

void  YumIO::collectStart (long long  totalFiles,
                           long long  totalSize,
                           long long  totalDrpms)
{
#if defined(DEBUG_COLLECT_DOWNLOAD)
   printf("collect: start: totalFiles=%llu, totalSize=%llu, totalDrpms=%llu\n",
          totalFiles, totalSize, totalDrpms);
#endif

   // This callback contains no actual information.  The paramters contain dummy
   // values.  So don't actually start until we see the first
   // downloadProgress call.
}

void  YumIO::collectProgress (QString          name,
                              Progress::Steps  size,
                              Progress::Steps  itemProgress)
{
#if defined(DEBUG_COLLECT_DOWNLOAD)
   printf("collect: %llu (payload=%s, downloadSize=%llu)\n",
          itemProgress, name.toStdString().c_str(), size);
#endif

   if (name != repoProgress.name) {
      repoProgress.name = name;
      repoProgress.act++;
      repoProgress.total = 1;

#if defined(DEBUG_COLLECT_DOWNLOAD_COOKED)
      printf("collect: Start name=%s, progress=%d/%d\n",
             repoProgress.name.toStdString().c_str(),
             repoProgress.act, repoProgress.repos.count());
#endif

      QString activity = tr("Reading repository metadata for %1")
                            .arg(repoProgress.name);
      pg->startItemAct(activity, repoProgress.total, repoProgress.act);
   } else if (itemProgress > 0) {
      // It's possible for the "finish line" to move at any moment in these
      // callbacks, so always just call changeItemActSteps instead of
      // setItemValue.

#if defined(DEBUG_COLLECT_DOWNLOAD_COOKED)
      printf("collect: Update: %llu/%u %s\n",
             itemProgress, repoProgress.total,
             repoProgress.name.toStdString().c_str());
#endif

      repoProgress.total = size;
      pg->changeItemActSteps(itemProgress, repoProgress.total);
   }
}

void  YumIO::collectMessage (QString  msg)
{
#if defined(DEBUG_COLLECT_DOWNLOAD)
   printf("collect: message=%s\n", msg.toStdString().c_str());
#endif

   // I've never seen this called, so I'm not sure what it does.
}

void  YumIO::collectEnd (QString          name,
                         Progress::Steps  size,
                         long             status,
                         QString          msg)
{
#if defined(DEBUG_COLLECT_DOWNLOAD)
   printf("collect: end: status=%ld msg=%s (payload=%s, downloadSize=%llu)\n",
          status, msg.toStdString().c_str(),
          name.toStdString().c_str(), size);
#endif

#if defined(DEBUG_COLLECT_DOWNLOAD_COOKED)
   printf("collect: End: %u/%u %s\n",
          repoProgress.total, repoProgress.total,
          repoProgress.name.toStdString().c_str());
#endif

   // Bump the act's progress to its "finish line", in case we never received
   // an update saying that.
   
   pg->setItemValue(repoProgress.total);
}

PyObject*  YumIO::downloadStart (PyObject*  self, PyObject*  args)
{
   long long  totalFiles;
   long long  totalSize;
   long long  totalDrpms;

   if (!PyArg_ParseTuple(args, "LLL", 
                         &totalFiles, &totalSize, &totalDrpms)) {
      if (PyErr_Occurred()) handleException();
      error(tr("Internal error: downloadStart: unable to interpret arguments"));
      return NULL;
   }

   switch (download.mode) {
   case Collect:
      collectStart(totalFiles, totalSize, totalDrpms);
      break;
   case Ts:
      tsDownloadStart(totalFiles, totalSize, totalDrpms);
      break;
   default:
      error(tr("Internal error: downloadStart: download at unexpected time"));
      return NULL;
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::downloadProgress (PyObject*  self, PyObject*  args)
{
   PyObject*  payload;
   double     done;
   
   if (!PyArg_ParseTuple(args, "Od", &payload, &done)) {
      if (PyErr_Occurred()) handleException();
      error(tr("Internal error: downloadProgress:"
               " unable to interpret arguments"));
      return NULL;
   }

   PyObject*  payloadNameObj = pyMethodCall(payload, "__str__", "()");
   if (payloadNameObj == NULL) {
      if (PyErr_Occurred()) handleException();
      error(tr("Internal error: downloadProgress:"
               " unable to determine payload __str__"));
      return NULL;
   }
   QString  payloadName = pyString(payloadNameObj);
   Py_XDECREF(payloadNameObj);

   PyObject*  payloadSizeObj = PyObject_GetAttrString(payload, "download_size");
   if (payloadSizeObj == NULL) {
      if (PyErr_Occurred()) handleException();
      error(tr("Internal error: downloadProgress:"
               " unable to determine payload download_size"));
      return NULL;
   }
   long  payloadSize = PY_TO_LONG(payloadSizeObj);
   Py_XDECREF(payloadSizeObj);

   switch (download.mode) {
   case Collect:
      collectProgress(payloadName, payloadSize, (Progress::Steps)done);
      break;
   case Ts:
      tsDownloadProgress(payloadName, payloadSize, (Progress::Steps)done);
      break;
   default:
      error(tr("Internal error: downloadProgress: download at unexpected time"));
      return NULL;
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::downloadMessage (PyObject*  self, PyObject*  args)
{
   PyObject*  msgObj;
   
   if (!PyArg_ParseTuple(args, "O", &msgObj)) {
      if (PyErr_Occurred()) handleException();
      error(tr("Internal error: downloadMessage:"
               " unable to interpret arguments"));
      return NULL;
   }

   QString  msg = pyString(msgObj);

   switch (download.mode) {
   case Collect:
      collectMessage(msg);
      break;
   case Ts:
      tsDownloadMessage(msg);
      break;
   default:
      error(tr("Internal error: downloadMessage: download at unexpected time"));
      return NULL;
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::downloadEnd (PyObject*  self, PyObject*  args)
{
   PyObject*  payload;
   PyObject*  statusObj;
   PyObject*  msgObj;
   
   if (!PyArg_ParseTuple(args, "OOO",
                         &payload, &statusObj, &msgObj)) {
      if (PyErr_Occurred()) handleException();
      error(tr("Internal error: downloadEnd:"
               " unable to interpret arguments"));
      return NULL;
   }

   PyObject*  payloadNameObj = pyMethodCall(payload, "__str__", "()");
   if (payloadNameObj == NULL) {
      if (PyErr_Occurred()) handleException();
      error(tr("Internal error: downloadEnd:"
               " unable to determine payload __str__"));
      return NULL;
   }
   QString  payloadName = pyString(payloadNameObj);
   Py_XDECREF(payloadNameObj);

   PyObject*  payloadSizeObj = PyObject_GetAttrString(payload, "download_size");
   if (payloadSizeObj == NULL) {
      if (PyErr_Occurred()) handleException();
      error(tr("Internal error: downloadEnd:"
               " unable to determine payload download_size"));
      return NULL;
   }
   long  payloadSize UNUSED = PY_TO_LONG(payloadSizeObj);
   Py_XDECREF(payloadSizeObj);

   long     status = (statusObj == Py_None ? 0 : PY_TO_LONG(statusObj));
   QString  msg    = pyString(msgObj);

   switch (download.mode) {
   case Collect:
      collectEnd(payloadName, payloadSize, status, msg);
      break;
   case Ts:
      tsDownloadEnd(payloadName, payloadSize, status, msg);
      break;
   default:
      error(tr("Internal error: downloadEnd: download at unexpected time"));
      return NULL;
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::userconfirm(PyObject*  self, PyObject*  args)
{ 
#if defined(DEBUG_TS)
   dprintf("userconfirm\n");
#endif

#if defined(DEBUG_TS_DESCR)
   dprintf("ts.descr: -----display-----\n");
#endif

   pg->reset(); // PG_D_*

   ts.act = 0;

   int  result;
   if (ts.list.during) {
      result = dialogs->confirm(tr("Apply"), ts.descr, "OK", "Cancel");
   } else {
      result = dialogs->confirm(tr("Query"), ts.descr, "Yes", "No");
   }

   bool  accepted = (result == QDialog::Accepted);

   if (ts.list.during) {
      if (accepted) {
         log(Log::ACTION, 0, tr("Applying: OK"));
      } else {
         log(Log::ACTION, 0, tr("Not Applying: Canceled"));
      }

      ts.list.confirmed = accepted;

      if (ts.list.confirmed) {
         // The PG_R_GRAB_RPMS_NOMINAL fractional part of PG_R_TOTAL is based
         // on actually having to grab all the rpm's.  If some already are in
         // the cache, then scale it back accordinly.
         ts.run.pgRGrabRpms = (ts.run.nominalGrabSize == 0
                               ? 0
                               : (ts.run.totalGrabSize * PG_R_GRAB_RPMS_NOMINAL
                                  / ts.run.nominalGrabSize));

#if defined(DEBUG_TS)
         dprintf("pgRGrabRpms = %u/%u\n",
                 ts.run.pgRGrabRpms, PG_R_GRAB_RPMS_NOMINAL);
#endif

         pg->startTotal(tr(PG_R_TOTAL_TEXT), PG_R_TOTAL);

         pg->startItem(tr(PG_R_INITIALIZE_TEXT), PG_R_INITIALIZE);

         // All rpms are not the same size, so use size as the number of acts
         pg->setupItemActs(ts.run.pgRGrabRpms, ts.run.totalGrabSize);
      }
   } else {
      if (accepted) {
         log(Log::ACTION, 0, tr("Confirmation: Yes"));
      } else {
         log(Log::ACTION, 0, tr("Confirmation: No"));
      }
   }

#if defined(DEBUG_TS_DESCR)
   dprintf("ts.descr: -----clear-----\n");
#endif

   ts.descr = QString();
   ts.list.during = false;

#if PY_VERSION_HEX >= 0x03000000
   return PyLong_FromLong(accepted);
#else
   return PyInt_FromLong(accepted);
#endif
}

PyObject*  YumIO::reportDownloadSize(PyObject*  self, PyObject*  args)
{ 
   PyObject*  packages;
   
   if (!PyArg_ParseTuple(args, "O", &packages)) return NULL;

#if defined(DEBUG_TS)
   dprintf("reportDownloadSize\n");
#endif

   Package::Size  localSize     = 0;
   Package::Size  downloadSize  = 0;
   Package::Size  installedSize = 0;

   PyObject*  it = PyObject_GetIter(packages);
   for (;;) {
      PyObject*  pkg = PyIter_Next(it);
      if (pkg == NULL) break;
      if (hasLocalPkg(pkg)) {
         localSize += sizeOf(pkg);
      } else {
         downloadSize += sizeOf(pkg);
      }
      installedSize += PY_TO_LONG(PyObject_GetAttrString(pkg, "installsize"));
      Py_DECREF(pkg);
   }
   
   Py_DECREF(it);

   ts.descr.append("\n");
   ts.descr.append(tr("Cached size: %1\n").arg(bytesFormat(localSize)));
   ts.descr.append(tr("Total download size: %1\n")
                      .arg(bytesFormat(downloadSize)));
   ts.descr.append(tr("Installed size: %1\n")
                      .arg(bytesFormat(installedSize)));

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::reportRemoveSize(PyObject*  self, PyObject*  args)
{ 
   PyObject*  packages;
   
   if (!PyArg_ParseTuple(args, "O", &packages)) return NULL;

#if defined(DEBUG_TS)
   dprintf("reportRemoveSize\n");
#endif

   ts.descr.append(tr("No downloads; only removals\n"));

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::listTransaction(PyObject*  self, PyObject*  args)
{ 
   PyObject*  output;
   PyObject*  transaction;
   
   if (!PyArg_ParseTuple(args, "OO", &output, &transaction)) return NULL;

#if defined(DEBUG_TS)
   dprintf("listTransaction\n");
#endif

#if defined(DEBUG_TS_DESCR)
   dprintf("ts.descr: -----clear-----\n");
#endif

   ts.descr = QString();
   ts.list.during = true;

   PyObject*  listBunch;
   PyObject*  downgraded;
   PyObject*  erased;
   PyObject*  erased_clean;
   PyObject*  erased_dep;
   PyObject*  installed;
   PyObject*  installed_group;
   PyObject*  installed_dep;
   PyObject*  installed_weak;
   PyObject*  reinstalled;
   PyObject*  upgraded;

   if (!makeLists(output, transaction, listBunch,
                  downgraded, erased, erased_clean, erased_dep,
                  installed, installed_group, installed_dep, installed_weak,
                  reinstalled, upgraded)) {
      return NULL;
   }

   ts.descr.append(tr("Transaction to be performed:\n"));

   if (   !showTransPkgsRepo(tr("Install"),         installed)
       || !showTransPkgsRepo(tr("Upgrade"),         upgraded)
       || !showTransPkgsRepo(tr("Reinstall"),       reinstalled)
       || !showTransPkgsRepo(tr("Install (group)"), installed_group)
       || !showTransPkgs    (tr("Erase"),           erased)
       || !showTransPkgsRepo(tr("Downgrade"),       downgraded)) {
      Py_DECREF(listBunch);
      return NULL;
   }

   if (   PySequence_Length(installed_dep) > 0
       || PySequence_Length(installed_weak) > 0
       || PySequence_Length(erased_dep) > 0
       || PySequence_Length(erased_clean) > 0) {
      ts.descr.append(tr("Performing the following to resolve"
                         " dependencies:\n"));
      if (   !showTransPkgsRepo(tr("Install"),         installed_dep)
          || !showTransPkgsRepo(tr("Install (weak)"),  installed_weak)
          || !showTransPkgs    (tr("Remove"),          erased_dep)
          || !showTransPkgs    (tr("Remove (unused)"), erased_clean)) {
         Py_DECREF(listBunch);
         return NULL;
      }
   }

   noteTransPkgs(Package::INSTALL, installed);
   noteTransPkgs(Package::INSTALL, installed_group);
   noteTransPkgs(Package::UPDATE,  upgraded);
   noteTransPkgs(Package::UPDATE,  reinstalled);
   noteTransPkgs(Package::UPDATE,  downgraded);
   noteTransPkgs(Package::REMOVE,  erased);

   noteTransPkgs(Package::INSTALL, installed_dep);
   noteTransPkgs(Package::INSTALL, installed_weak);
   noteTransPkgs(Package::REMOVE,  erased_dep);
   noteTransPkgs(Package::REMOVE,  erased_clean);

   Py_DECREF(listBunch);

   Py_INCREF(Py_None);
   return Py_None;
}

void  YumIO::tsDownloadStart (long long  totalFiles,
                              long long  totalSize,
                              long long  totalDrpms)
{
#if defined(DEBUG_TS_DOWNLOAD)
   printf("tsDownload: start:"
          " totalFiles=%llu, totalSize=%llu, totalDrpms=%llu\n",
          totalFiles, totalSize, totalDrpms);
#endif

#if defined(DEBUG_TS_DOWNLOAD_COOKED)
   printf("tsDownload: Multi start: totalFiles=%llu totalSize=%llu\n",
          totalFiles, totalSize);
#endif

   // We don't know which activities are involved in this parallel block
   // of acts until we see the first applyGrabMeterStart.  So, just flag
   // that it has begun, and let applyGrabMeterStart() do the 
   // pg->setItemActsThread when it sees the first activity.
   ts.parallel.begun          = true;
   ts.parallel.files          = totalFiles;
   ts.parallel.size           = totalSize;
   ts.parallel.threadsStarted = 0;
   ts.parallel.meterMapping.clear();

   QString  activity = QString("Downloading %1 packages").arg(ts.parallel.files);
   Progress::Steps  firstAct = ts.act;
   ts.act += PG_R_GRAB(totalSize);
   Progress::Steps  tailAct = ts.act;
   pg->setupItemActsThreads(activity,
                            firstAct,
                            tailAct,
                            ts.parallel.files,
                            ts.parallel.size);
}

void  YumIO::tsDownloadProgress (QString          name,
                                 Progress::Steps  size,
                                 Progress::Steps  itemProgress)
{
#if defined(DEBUG_TS_DOWNLOAD)
   printf("tsDownload: %llu (payload=%s, downloadSize=%llu)\n",
          itemProgress, name.toStdString().c_str(), size);
#endif

   if (itemProgress == 0) {
      // We see lots of 0 progress updates early.  Don't treat them as having
      // started until there's at least *some* progress.
      // DO NOTHING
   } else {
      TsMeterMapping::iterator  meter = ts.parallel.meterMapping.find(name);
      if (meter == ts.parallel.meterMapping.end()) {
         // First non-0 progress seen for package; create new meter
         Progress::Threads  thread = ts.parallel.threadsStarted++;
         ts.parallel.meterMapping[name] = thread;

#if defined(DEBUG_TS_DOWNLOAD_COOKED)
         printf("tsDownload: start: %llu/%llu %s\n",
                itemProgress, size, name.toStdString().c_str());
#endif

         QString  activity = tr("Downloading %1").arg(name);
         pg->startItemActThread(activity, thread, size);
         pg->setItemActsThreadValue(thread, itemProgress);
      } else {
         Progress::Threads&  thread(*meter);

#if defined(DEBUG_TS_DOWNLOAD_COOKED)
         printf("tsDownload: %llu/%llu %s\n",
                itemProgress, size, name.toStdString().c_str());
#endif
         pg->setItemActsThreadValue(thread, itemProgress);
      }
   }
}

void  YumIO::tsDownloadMessage (QString  msg)
{
#if defined(DEBUG_TS_DOWNLOAD)
   printf("tsDownload: message=%s\n", msg.toStdString().c_str());
#endif

   // I've never seen this called, so I'm not sure what it does.
}

void  YumIO::tsDownloadEnd (QString          name,
                            Progress::Steps  size,
                            long             status,
                            QString          msg)
{
#if defined(DEBUG_TS_DOWNLOAD)
   printf("tsDownload: end: status=%ld msg=%s (payload=%s, downloadSize=%llu)\n",
          status, msg.toStdString().c_str(),
          name.toStdString().c_str(), size);
#endif

#if defined(DEBUG_TS_DOWNLOAD_COOKED)
   printf("tsDownload: end: %llu/%llu %s\n",
          size, size, name.toStdString().c_str());
#endif

   Progress::Threads  thread = ts.parallel.meterMapping[name];
   pg->setItemActsThreadValue(thread, size);
   pg->endItemActThread(thread);

   ts.parallel.meterMapping.remove(name);
}

PyObject*  YumIO::rpmtransInit (PyObject*  self, PyObject*  args)
{
   PyObject*        testObj;
   
   if (!PyArg_ParseTuple(args, "O", &testObj)) {
      if (PyErr_Occurred()) handleException();
      error(tr("Internal error: rpmtransInit: unable to interpret arguments"));
      return NULL;
   }

   bool  test = (testObj != Py_False);

#if defined(DEBUG_TS_DETAIL)
   printf("rpmtransInit (test=%s)\n", (test ? "True" : "False"));
#endif

   if (test || !ts.run.doingTest) {
      // The download* interfaces, and thus the tsDownload* have no callback
      // for completion of all downloads.  So do that here.
      //
      // Properly notice this when called with test=True.  But I don't trust dnf
      // to always provide a callback with test=True.  So also notice this if
      // called with test=False if there wasn't a prior call with test=True.
      // (That also would mean that the PG_R_RUN_TRANS_TEST would start and end
      // at the same time, but progress meter was budgeted for that phase, so it
      // has to happen at some point.)
      pg->endItemActsThreads();

      ts.parallel.begun          = false;
      ts.parallel.threadsStarted = 0;
      ts.parallel.meterMapping.clear();

      pg->endItemActs(); // for PG_R_GRAB_RPMS

      pg->startItem(tr(PG_R_RUN_TRANS_TEST_TEXT), PG_R_RUN_TRANS_TEST);
   }

   if (!test) {
      pg->setupItemActs(PG_R_RUN_TRANS, ts.run.totalTsSize);
   }
   
   ts.act           = 0;
   ts.run.doingTest = test;
   
   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::rpmtransDel (PyObject*  self, PyObject*  args)
{ 
#if defined(DEBUG_TS_DETAIL)
   printf("rpmtransDel\n");
#endif
   
   if (ts.run.doingTest) {
      // completes PG_R_RUN_TRANS_TEST, but that was a simple Item with no acts.
   } else {
      pg->endItemActs(); // for PG_R_RUN_TRANS
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::rpmtransProgress (PyObject*  self, PyObject*  args)
{
   PyObject*        pkg;
   long             action;
   PyObject*        progressObj;
   PyObject*        totalObj;
   PyObject*        currentTsObj;
   PyObject*        totalTsObj;
   bool             understood = false;
   
   if (!PyArg_ParseTuple(args, "OlOOOO",
                         &pkg, &action,
                         &progressObj, &totalObj,
                         &currentTsObj, &totalTsObj)) {
      if (PyErr_Occurred()) handleException();
      error(tr("Internal error: rpmtransProgress:"
               " unable to interpret arguments"));
      return NULL;
   }

#if defined(DEBUG_TS_DETAIL)
   QString  actStr;
   if      (action == PKG_INSTALL)        actStr = "PKG_INSTALL";
   else if (action == PKG_DOWNGRADE)      actStr = "PKG_DOWNGRADE";
   else if (action == PKG_OBSOLETE)       actStr = "PKG_OBSOLETED";
   else if (action == PKG_UPGRADE)        actStr = "PKG_UPGRADE";
   else if (action == PKG_REMOVE)         actStr = "PKG_REMOVE";
   else if (action == PKG_REINSTALL)      actStr = "PKG_REINSTALL";
   else if (action == PKG_CLEANUP)        actStr = "PKG_CLEANUP";
   else if (action == PKG_VERIFY)         actStr = "PKG_VERIFY";
   else if (action == PKG_SCRIPTLET)      actStr = "PKG_SCRIPTLET";
   else if (action == TRANS_PREPARATION)  actStr = "TRANS_PREPARATION";
   else if (action == TRANS_POST)         actStr = "TRANS_POST";
   else                                   actStr = QString::number(action);
#endif
   
   Progress::Steps  progress       = (progressObj == Py_None
                                      ? 0
                                      : PyLong_AsLongLong(progressObj));
   Progress::Steps  total          = (totalObj == Py_None
                                      ? 0
                                      : PyLong_AsLongLong(totalObj));
   unsigned long    currentTs      = (currentTsObj == Py_None
                                      ? 0
                                      : PY_TO_LONG(currentTsObj));
   unsigned long    totalTs UNUSED = (totalTsObj == Py_None
                                      ? 0
                                      : PY_TO_LONG(totalTsObj));

#if defined(DEBUG_TS_DETAIL)
   printf("rpmtransProgress (action=%s, progress=%llu/%llu, ts=%lu/%lu)\n",
          actStr.toStdString().c_str(),
          progress, total, currentTs, totalTs);
#endif

   if (action == TRANS_PREPARATION || action == TRANS_POST) {
      understood = true;
      // DO NOTHING
   } else if (action == PKG_INSTALL
              || action == PKG_DOWNGRADE
              || action == PKG_UPGRADE
              || action == PKG_REINSTALL
              || action == PKG_REMOVE
              || action == PKG_OBSOLETE
              || action == PKG_SCRIPTLET) {
      // The currentTs is incremented when moving to a new package.
      if (currentTs != ts.run.lastTs || action != ts.run.lastAction) {
         QString  name = pyStringDec(PyObject_GetAttrString(pkg, "name"));
         QString  arch = pyStringDec(PyObject_GetAttrString(pkg, "arch"));
#if defined(DEBUG_TS_DETAIL)
         printf("                 (name=%s) (arch=%s)\n",
                name.toStdString().c_str(),
                arch.toStdString().c_str());
#endif
         Package::Entry*  entry = hash[EntryKey(false, name, getBiArch(arch))];
         QString          activity;
         if (action == PKG_UPGRADE) {
            activity = tr("Upgrading %1").arg(name);
            ts.act += PG_R_UPDATE(entry->size);
         } else if (action == PKG_DOWNGRADE) {
            activity = tr("Downgrading %1").arg(name);
            ts.act += PG_R_UPDATE(entry->size);
         } else if (action == PKG_REINSTALL) {
            activity = tr("Reinstalling %1").arg(name);
            ts.act += PG_R_UPDATE(entry->size);
         } else if (action == PKG_INSTALL) {
            activity = tr("Installing %1").arg(name);
            ts.act += PG_R_INSTALL(entry->size);
         } else if (action == PKG_OBSOLETE) {
            activity = tr("Obsoleting %1").arg(name);
            ts.act += PG_R_REMOVE(entry->size);
         } else if (action == PKG_REMOVE) {
            activity = tr("Removing %1").arg(name);
            ts.act += PG_R_REMOVE(entry->size);
         } else if (action == PKG_SCRIPTLET) {
            activity = tr("Executing scriptlet for %1").arg(name);
            // TODO: Define some activity time for scriptlets.  Use 0 for now.
         } else {
            warning(tr("Internal error: YumIO::rpmtransProgress"
                       " interior action=%1 not understood").arg(action));
         }
            
         log(Log::NORMAL, 1, activity);
         pg->startItemAct(activity, total, ts.act);
         ts.run.lastTs     = currentTs;
         ts.run.lastAction = action;
      }
      pg->setItemValue(progress);
      understood = true;
   } else if (action == PKG_CLEANUP) {
      // DO NOTHING
      understood = true;
   } else if (action == PKG_VERIFY) {
      // DO NOTHING
      understood = true;
   }
      
   if (!understood) {
      warning(tr("Internal error: YumIO::rpmtransProgress"
                 " action=%1 not understood; ignoring").arg(action));
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::rpmtransScriptout (PyObject*  self, PyObject*  args)
{
   PyObject*  msgsObj;

   if (!PyArg_ParseTuple(args, "O", &msgsObj)) {
      if (PyErr_Occurred()) handleException();
      error(tr("Internal error: rpmtransScriptout:"
               " unable to interpret arguments"));
      return NULL;
   }

   if (msgsObj != Py_None) {
      QString  msgs = pyString(msgsObj);
      log(Log::STDOUT, 0, msgs);
   }

   Py_INCREF(Py_None);
   return Py_None;
}

PyObject*  YumIO::postTransactionOutput(PyObject*  self, PyObject*  args)
{ 
   PyObject*  output;
   PyObject*  transaction;
   
   if (!PyArg_ParseTuple(args, "OO", &output, &transaction)) return NULL;

   PyObject*  listBunch;
   PyObject*  downgraded;
   PyObject*  erased;
   PyObject*  erased_clean;
   PyObject*  erased_dep;
   PyObject*  installed;
   PyObject*  installed_group;
   PyObject*  installed_dep;
   PyObject*  installed_weak;
   PyObject*  reinstalled;
   PyObject*  upgraded;

   if (!makeLists(output, transaction, listBunch,
                  downgraded, erased, erased_clean, erased_dep,
                  installed, installed_group, installed_dep, installed_weak,
                  reinstalled, upgraded)) {
      return NULL;
   }

#if defined(DEBUG_TS)
   dprintf("postTransactionOutput\n");
#endif

   if (   !showTransPkgs(tr("Erased"),                      erased)
       || !showTransPkgs(tr("Dependency Erased"),           erased_dep)
       || !showTransPkgs(tr("Dependency Erased (Unused)"),  erased_clean)
       || !showTransPkgs(tr("Installed"),                   installed)
       || !showTransPkgs(tr("Installed (Group)"),           installed_group)
       || !showTransPkgs(tr("Dependency Installed"),        installed_dep)
       || !showTransPkgs(tr("Dependency Installed (Weak)"), installed_weak)
       || !showTransPkgs(tr("Upgraded"),                    upgraded)
       || !showTransPkgs(tr("Downgraded"),                  downgraded)
       || !showTransPkgs(tr("Reinstalled"),                 reinstalled)) {
      Py_DECREF(listBunch);
      return NULL;
   }

   Py_DECREF(listBunch);

#if PY_VERSION_HEX >= 0x03000000
   return PyUnicode_FromString("");
#else
   return PyString_FromString("");
#endif
}

PyObject*  YumIO::gpgsigcheck(PyObject*  self, PyObject*  args)
{
#if defined(DEBUG_TS)
   dprintf("gpgsigcheck\n");
#endif

   // This may be leading to a GPG key confirmation.  Clear out the accumulated
   // output so we have an uncluttered dialog.

#if defined(DEBUG_TS_DESCR)
   dprintf("ts.descr: -----clear-----\n");
#endif

   ts.descr = QString();

   Py_INCREF(Py_None);
   return Py_None;
}

/*
** Utility functions
*/

QString  YumIO::getLocalPkg (PyObject*  pkg)
{
   PyObject*  obj = pyMethodCall(pkg, "localPkg", "()");
   if (obj == NULL || obj == Py_None) return QString();
   return pyStringDec(obj);
}

bool  YumIO::hasLocalPkg (PyObject*  pkg)
{
   pyExceptStart();
   pyExceptExpect(PyExc_FileNotFoundError);
   pyExceptExpect(xcpt.miscError);
   PyObject*  obj = pyMethodCall(pkg, "verifyLocalPkg", "()");
   if (pyExceptMatches(obj, PyExc_FileNotFoundError)
       || pyExceptMatches(obj, xcpt.miscError)) {
      pyExceptClear();
      return false;
   }
   pyExceptClear();
   bool  result = (obj != Py_None);
   Py_XDECREF(obj);
   return result;
}

Package::Kind  YumIO::getPkgKind (PyObject*  pkg)
{
   return simpleString(pkg, "group") == "Suite" ? Package::SUITE 
                                                : Package::PACKAGE;
}

PyObject*  YumIO::getGroupItem_22 (const Package::Entry&  entry,
                                   char*                  attr)
{
   if (hasDistribGroups()) {
      error(tr("getGroupItem_22 called for incorrect version of yum"));
      return NULL;
   }

   PyObject*  dict = PyObject_GetAttrString(entry.obj, attr);
   if (dict == NULL) {
      error(tr("Internal error: groupInfo.%1 not evaluable").arg(attr));
      return NULL;
   }
   PyObject*  result = pyDictGetItem(dict, entry.name);
   if (result == NULL) {
      error(tr("Internal error: groupInfo.%1[%2]"
               " not evaluable").arg(attr).arg(entry.name));
      return NULL;
   }
   Py_INCREF(result); // return a new reference to be like getGroupItem_25
   Py_DECREF(dict);
   return result;
}

PyObject*  YumIO::getGroupItem_25 (const Package::Entry&  entry,
                                   char*                  attr)
{
   if (!hasDistribGroups()) {
      error(tr("getGroupItem_25 called for incorrect version of yum"));
      return NULL;
   }
      
   PyObject*  result = PyObject_GetAttrString(entry.obj, attr);
   if (result == NULL) {
      error(tr("Internal error: group %1.%2 not evaluable").arg(attr)
                                                           .arg(entry.name));
      return NULL;
   }
   return result;
}

Package::Entry*  YumIO::findPackage (QString  name)
{
   // Find a package given only the package name and no architecture information
   // (e.g. from a package group's list of packages).  First try the package for
   // the basearch.  If not found, pick any package with that name by checking
   // each architecture ever seen in the latest collect pass.

   EntryKey  key(false, name, getBiArch(basearch));
   if (hash.contains(key)) {
      return hash[key];
   }

   StringSet::const_iterator  it = biarches.begin();
   for (; it != biarches.end(); ++it) {
      if (*it != basearch) {
         EntryKey  key(false, name, *it);
         if (hash.contains(key)) {
            return hash[key];
         }
      }
   }

   return NULL;
}

bool  YumIO::parseEnvra (QString   envra,
                         QString&  name,
                         QString&  epoch,
                         QString&  version,
                         QString&  release,
                         QString&  arch)
{
   // The envra is in this format: [epoch:]name-version-release.arch

   int  beforeArch = envra.lastIndexOf('.');
   if (beforeArch <= 0) return false;

   int  beforeRel = envra.lastIndexOf('-', beforeArch-1);
   if (beforeRel <= 0) return false;

   int  beforeVer = envra.lastIndexOf('-', beforeRel-1);
   if (beforeVer <= 0) return false;

   int  beforeName = envra.lastIndexOf(':', beforeVer-1);
   // beforeName can be -1 legimitately

   if (beforeName == -1) {
      epoch = QString();
      // beforeName = -1 // coincidentally, it already is -1!
   } else {
      epoch = envra.mid(0, beforeName);
   }
   name    = envra.mid(beforeName+1, beforeVer - beforeName - 1);
   version = envra.mid(beforeVer+1, beforeRel - beforeVer - 1);
   release = envra.mid(beforeRel+1, beforeArch - beforeRel - 1);
   arch    = envra.mid(beforeArch+1);

#ifdef DEBUG_ENVRA
   printf("envra %s -> (n=%s, e=%s, v=%s, r=%s, a=%s)\n",
          envra.toStdString().c_str(),
          name.toStdString().c_str(),
          epoch.toStdString().c_str(),
          version.toStdString().c_str(),
          release.toStdString().c_str(),
          arch.toStdString().c_str());
#endif

   return true;
}

bool  YumIO::parseEnvra (QString   envra,
                         QString&  name,
                         QString&  arch)
{
   QString  epoch;
   QString  version;
   QString  release;
   return parseEnvra(envra, name, epoch, version, release, arch);
}

bool  YumIO::applyGroups()
{
   Package::EntryList::const_iterator  it = list.begin();
   for (; it != list.end(); ++it) {
      Package::Entry*  entry = *it;

      if (entry->kind == Package::GROUP) {
         switch (entry->apply.action) {
         case Package::UPDATE:
            if (!applyGrpUpdate(entry)) return false;
            break;
         case Package::INSTALL:
            if (!applyGrpInstall(entry)) return false;
            break;
         case Package::REMOVE:
            if (!applyGrpRemove(entry)) return false;
            break;
         case Package::NONE:
            // DO NOTHING
            break;
         }
      }
   }

   return true;
}

bool  YumIO::applyGrpUpdate (const Package::Entry*  grp)
{
   QSet<QString>  reqs;
   {
      PyObject*  reqPkgs = requiredPkgs(grp);
      if (reqPkgs == NULL) return false;
      PyObject*  it      = PyObject_GetIter(reqPkgs);
      for (;;) {
         PyObject*  pkgItem = PyIter_Next(it);
         if (pkgItem == NULL) break;
         reqs.insert(pyString(pkgItem));
         Py_DECREF(pkgItem);
      }
      Py_DECREF(it);
      Py_DECREF(reqPkgs);
   }
   
   PyObject*  pkgs = pkgTree(grp);
   if (pkgs == NULL) return false;
   PyObject*  it   = PyObject_GetIter(pkgs);
   for (;;) {
      PyObject*  pkgItem = PyIter_Next(it);
      if (pkgItem == NULL) break;

      Package::Entry*  entry = findPackage(pyString(pkgItem));
      if (entry != NULL) {
         switch (entry->apply.action) {
         case Package::NONE:
            // Only install required packages.  But if an optional package is
            // installed, do update it if necessary.
            switch (entry->state) {
            case Package::OUT_OF_DATE:
               entry->apply.action = Package::UPDATE;
               applyGrpLog(grp, entry);
               break;
            case Package::NOT_INSTALLED:
               if (reqs.contains(entry->name)) {
                  entry->apply.action = Package::INSTALL;
                  applyGrpLog(grp, entry);
               } else {
                  applyGrpLogNot(grp, entry, Package::INSTALL,
                                 tr("the package is optional"));
               }
               break;
            case Package::INSTALLED:
               // DO NOTHING
            default:
               // DO NOTHING
               break;
            }
            break;
         case Package::UPDATE:
         case Package::INSTALL:
            // DO NOTHING: no conflict
            applyGrpLogWas(grp, entry);
            break;
         case Package::REMOVE:
         default:
            if (!applyGrpConflict(grp, entry)) {
               return false;
            }
            break;
         }
      }

      Py_DECREF(pkgItem);
   }
   Py_DECREF(it);
   Py_DECREF(pkgs);

   return true;
}

bool  YumIO::applyGrpInstall (const Package::Entry*  grp)
{
   PyObject*  pkgs = pkgTree(grp);
   if (pkgs == NULL) return false;
   PyObject*  it   = PyObject_GetIter(pkgs);
   for (;;) {
      PyObject*  pkgItem = PyIter_Next(it);
      if (pkgItem == NULL) break;

      
      Package::Entry*  entry = findPackage(pyString(pkgItem));
      if (entry != NULL) {
         switch (entry->apply.action) {
         case Package::NONE:
            // Install or update all packages
            switch (entry->state) {
            case Package::OUT_OF_DATE:
               entry->apply.action = Package::UPDATE;
               applyGrpLog(grp, entry);
               break;
            case Package::NOT_INSTALLED:
               entry->apply.action = Package::INSTALL;
               applyGrpLog(grp, entry);
               break;
            case Package::INSTALLED:
            default:
               // DO NOTHING
               break;
            }
            break;
         case Package::UPDATE:
         case Package::INSTALL:
            // DO NOTHING: no conflicts
            applyGrpLogWas(grp, entry);
            break;
         case Package::REMOVE:
         default:
            if (!applyGrpConflict(grp, entry)) {
               return false;
            }
            break;
         }
      }

      Py_DECREF(pkgItem);
   }
   Py_DECREF(it);
   Py_DECREF(pkgs);

   return true;
}

bool  YumIO::applyGrpRemove (const Package::Entry*  grp)
{
   PyObject*  pkgs = allPkgs(grp);
   PyObject*  it   = PyObject_GetIter(pkgs);
   for (;;) {
      PyObject*  pkgItem = PyIter_Next(it);
      if (pkgItem == NULL) break;

      Package::Entry*  entry = findPackage(pyString(pkgItem));
      if (entry != NULL) {
         switch (entry->apply.action) {
         case Package::NONE:
            switch (entry->state) {
            case Package::OUT_OF_DATE:
            case Package::INSTALLED:
               entry->apply.action = Package::REMOVE;
               applyGrpLog(grp, entry);
               break;
            case Package::NOT_INSTALLED:
            default:
               // DO NOTHING
               break;
            }
            break;
         case Package::REMOVE:
            // DO NOTHING: no conflict
            applyGrpLog(grp, entry);
            break;
         case Package::UPDATE:
         case Package::INSTALL:
         default:
            if (!applyGrpConflict(grp, entry)) {
               return false;
            }
            break;
         }
      }

      Py_DECREF(pkgItem);
   }
   Py_DECREF(it);
   Py_DECREF(pkgs);

   return true;
}

bool  YumIO::applyGrpConflict (const Package::Entry*  grp,
                               Package::Entry*        pkg)
{
   Package::Action  grpPkgAction;

   switch (pkg->state) {
   case Package::OUT_OF_DATE:   grpPkgAction = Package::UPDATE;  break;
   case Package::NOT_INSTALLED: grpPkgAction = Package::INSTALL; break;
   case Package::INSTALLED:     grpPkgAction = Package::REMOVE;  break;
   default:
      error("Internal error: applyGrpConflict unexpected package state\n");
      return false;
   }

   QString  grpImage    = ACTION_IMAGE(grp->apply.action);
   QString  grpPkgImage = ACTION_IMAGE(grpPkgAction);
   QString  pkgImage    = ACTION_IMAGE(pkg->apply.action);

   QString  msg;
   if (pkg->apply.reason == pkg) {
      msg = tr("Package %1 is included in group %2 which was marked for %3,\n"
               "so the package would be marked for %4,\n"
               "but it was marked individually for %5.\n"
               "How do you wish to proceed?")
               .arg(pkg->name)
               .arg(grp->name)
               .arg(grpImage)
               .arg(grpPkgImage)
               .arg(pkgImage);
   } else if (pkg->apply.reason != NULL) {
      msg = tr("Package %1 is included in group %2 which was marked for %3,\n"
               "so the package would be marked for %4,\n"
               "but it also was part of group %5 and so was marked for %6.\n"
               "How do you wish to proceed?")
               .arg(pkg->name)
               .arg(grp->name)
               .arg(grpImage)
               .arg(grpPkgImage)
               .arg(pkg->apply.reason->name)
               .arg(pkgImage);
   } else {
      msg = tr("Package %1 is included in group %2 which was marked for %3,\n"
               "so the package would be marked for %4,\n"
               "but you resolved an earlier conflict by marking it for %5.\n"
               "How do you wish to proceed?")
               .arg(pkg->name)
               .arg(grp->name)
               .arg(grpImage)
               .arg(grpPkgImage)
               .arg(pkgImage);
   }

   int  result = dialogs->critical(/* caption             => */ tr("Apply"),
                                   /* text                => */ msg,
                                   /* button0Text         => */ grpPkgImage,
                                   /* button1Text         => */ pkgImage,
                                   /* button2Text         => */ tr("Cancel"),
                                   /* defaultButtonNumber => */ 2,
                                   /* escapeButtonNumber  => */ 2);
   
   switch (result) {
   case 0:
      pkg->apply.action = grpPkgAction;
      break;
   case 1:
      // DO NOTHING
      break;
   case 2:
      return false;
   }

   log(Log::NORMAL, 4, tr("Manually resolving conflict: marking package %1 for"
                          " %3").arg(pkg->name)
                                .arg(ACTION_IMAGE(pkg->apply.action)));
   
      
   pkg->apply.reason = NULL; // to indicate a user conflict resolution
   return true;
}

void  YumIO::applyGrpLog (const Package::Entry*  grp,
                          const Package::Entry*  pkg)
{
   log(Log::NORMAL, 4, tr("For group %1, marking package %2 for"
                          " %3").arg(grp->name)
                                .arg(pkg->name)
                                .arg(ACTION_IMAGE(pkg->apply.action)));
}

void  YumIO::applyGrpLogWas (const Package::Entry*  grp,
                             const Package::Entry*  pkg)
{
   log(Log::NORMAL, 4, tr("For group %1, package %2 already was marked for"
                          " %3").arg(grp->name)
                                .arg(pkg->name)
                                .arg(ACTION_IMAGE(pkg->apply.action)));
}

void  YumIO::applyGrpLogNot (const Package::Entry*  grp,
                             const Package::Entry*  pkg,
                             Package::Action        action,
                             QString                reason)
{
   log(Log::NORMAL, 4, tr("For group %1, not marking package %2 for"
                          " %3, becuase %4").arg(grp->name)
                                            .arg(pkg->name)
                                            .arg(ACTION_IMAGE(action))
                                            .arg(reason));
}

PyObject*  YumIO::requiredPkgs (const Package::Entry*  grp)
{
   if (hasDistribGroups()) {
      PyObject*  mand = PyObject_GetAttrString(grp->obj, "mandatory_packages");
      PyObject*  def  = PyObject_GetAttrString(grp->obj, "default_packages");
      if (mand == NULL || def == NULL) {
         error(tr("Internal error: group lacks required package lists"));
         return NULL;
      }
      PyObject*  mand_keys = pyMethodCall(mand, "keys", "()");
      PyObject*  def_keys  = pyMethodCall(def,  "keys", "()");
      if (mand_keys == NULL || def_keys == NULL) {
         error(tr("Internal error:"
                  " group lacks required proper required package lists"));
         return NULL;
      }
      PyObject*  result = PySequence_Concat(mand_keys, def_keys);
      Py_DECREF(def_keys);
      Py_DECREF(mand_keys);
      Py_DECREF(def);
      Py_DECREF(mand);
      return result;
   } else {
      char*      name   = strdup(grp->name.toStdString().c_str());
      PyObject*  result = pyMethodCall(grp->obj, "requiredPkgs", "(s)", name);
      free(name);
      return result;
   }
}

PyObject*  YumIO::pkgTree (const Package::Entry*  grp)
{
   if (hasDistribGroups()) {
      // There are no subgroups in yum >= 2.5, so these two are equivalent.
      return requiredPkgs(grp);
   } else {
      char*      name   = strdup(grp->name.toStdString().c_str());
      PyObject*  result = pyMethodCall(grp->obj, "pkgTree", "(s)", name);
      free(name);
      return result;
   }
}

PyObject*  YumIO::allPkgs (const Package::Entry*  grp)
{
   if (hasDistribGroups()) {
      PyObject*  mand = PyObject_GetAttrString(grp->obj, "mandatory_packages");
      PyObject*  def  = PyObject_GetAttrString(grp->obj, "default_packages");
      PyObject*  opt  = PyObject_GetAttrString(grp->obj, "optional_packages");
      PyObject*  cond = PyObject_GetAttrString(grp->obj, 
                                               "conditional_packages");
      if (mand == NULL || def == NULL || opt == NULL || cond == NULL) {
         error(tr("Internal error: group lacks package lists"));
         return NULL;
      }
      PyObject*  mand_keys = pyMethodCall(mand, "keys", "()");
      PyObject*  def_keys  = pyMethodCall(def,  "keys", "()");
      PyObject*  opt_keys  = pyMethodCall(opt,  "keys", "()");
      PyObject*  cond_keys = pyMethodCall(cond, "keys", "()");
      if (mand_keys == NULL || def_keys == NULL
          || opt_keys == NULL || cond_keys == NULL) {
         error(tr("Internal error:"
                  " group lacks required proper package lists"));
         return NULL;
      }
      PyObject*  temp1  = PySequence_Concat(mand_keys, def_keys);
      PyObject*  temp2  = PySequence_Concat(opt_keys,  cond_keys);
      PyObject*  result = PySequence_Concat(temp1, temp2);
      Py_DECREF(temp2);
      Py_DECREF(temp1);
      Py_DECREF(cond_keys);
      Py_DECREF(opt_keys);
      Py_DECREF(def_keys);
      Py_DECREF(mand_keys);
      Py_DECREF(cond);
      Py_DECREF(opt);
      Py_DECREF(def);
      Py_DECREF(mand);
      return result;
   } else {
      char*      name   = strdup(grp->name.toStdString().c_str());
      PyObject*  result = pyMethodCall(grp->obj, "requiredPkgs", "(s)", name);
      free(name);
      return result;
   }
}

bool  YumIO::makeLists (PyObject*   output,
                        PyObject*   transaction,
                        PyObject*&  listBunch,
                        PyObject*&  downgraded,
                        PyObject*&  erased,
                        PyObject*&  erased_clean,
                        PyObject*&  erased_dep,
                        PyObject*&  installed,
                        PyObject*&  installed_group,
                        PyObject*&  installed_dep,
                        PyObject*&  installed_weak,
                        PyObject*&  reinstalled,
                        PyObject*&  upgraded)
{
   PyObject*  base = PyObject_GetAttrString(output, "base");
   if (base == NULL) {
      if (PyErr_Occurred()) handleException();
      error(tr("Internal error: cannot find output.base"));
      return false;
   }
   PyObject*  goal = PyObject_GetAttrString(base, "goal");
   if (goal == NULL) {
      if (PyErr_Occurred()) handleException();
      error(tr("Internal error: cannot find output.base.goal"));
      return false;
   }

   listBunch = pyCall("dnf.cli.output._make_lists", "(O,O)", transaction, goal);

   downgraded      = PyObject_GetAttrString(listBunch, "downgraded");
   erased          = PyObject_GetAttrString(listBunch, "erased");
   erased_clean    = PyObject_GetAttrString(listBunch, "erased_clean");
   erased_dep      = PyObject_GetAttrString(listBunch, "erased_dep");
   installed       = PyObject_GetAttrString(listBunch, "installed");
   installed_group = PyObject_GetAttrString(listBunch, "installed_group");
   installed_dep   = PyObject_GetAttrString(listBunch, "installed_dep");
   installed_weak  = PyObject_GetAttrString(listBunch, "installed_weak");
   reinstalled     = PyObject_GetAttrString(listBunch, "reinstalled");
   upgraded        = PyObject_GetAttrString(listBunch, "upgraded");
   // The only reference count held after return from this function is that for
   // listBunch.  Reference to each of the lists (e.g. installed) is held only
   // indirectly via the listBunch.  That simplifies things for the caller,
   // because it can free all of them with a single DECREF on the listBunch.
   Py_DECREF(downgraded);
   Py_DECREF(erased);
   Py_DECREF(erased_clean);
   Py_DECREF(erased_dep);
   Py_DECREF(installed);
   Py_DECREF(installed_dep);
   Py_DECREF(installed_weak);
   Py_DECREF(reinstalled);
   Py_DECREF(upgraded);

   Py_DECREF(goal);
   Py_DECREF(base);

   return true;
}

QString  YumIO::simpleString (PyObject*    pkg,
                              const char*  key)
{
   PyObject*  obj = PyObject_GetAttrString(pkg, key);
   if (obj == NULL || obj == Py_None) return QString();

   QString  result = pyStringXDec(obj);
   result.remove(QRegExp("\n*$"));

   return result;
}

QString  YumIO::getVerRel (PyObject*  pkg)
{
   QString    epoch;
   PyObject*  epochObj = PyObject_GetAttrString(pkg, "epoch");
   if (epochObj != NULL && PyLong_Check(epochObj)) {
      epoch = QString::number(PY_TO_LONG(epochObj));
   } else {
      epoch = pyString(epochObj);
   }
   Py_XDECREF(epochObj);
   
   QString  ver   = pyStringDec(PyObject_GetAttrString(pkg, "version"));
   QString  rel   = pyStringDec(PyObject_GetAttrString(pkg, "release"));

   QString  result;
   if (!epoch.isEmpty() && epoch != "0") {
      result.append(epoch).append(':');
   }
   result.append(ver);
   if (!rel.isEmpty()) {
      result.append('-').append(rel);
   }
   return result;
}

QString  YumIO::getVerRelTxmbr (PyObject*  txmbr)
{
   long     epoch = PyLong_AsLongLong(PyObject_GetAttrString(txmbr, "epoch"));
   QString  ver   = pyStringDec(PyObject_GetAttrString(txmbr, "version"));
   QString  rel   = pyStringDec(PyObject_GetAttrString(txmbr, "release"));

   QString  result;
   if (epoch != 0) {
      result.append(QString::number(epoch)).append(':');
   }
   result.append(ver);
   if (!rel.isEmpty()) {
      result.append('-').append(rel);
   }
   return result;
}

QString  YumIO::getBiArch(QString  arch)
{
   QString  result;

   char*  arch_s = strdup(arch.toStdString().c_str());
   result = pyStringDec(pyMethodCall(yumIO, "getBiArch", "(s)", arch_s));
   free(arch_s);

   return result;
}

Package::Size  YumIO::sizeOf(PyObject*  pkg)
{
   PyObject*      sizeObj = PyObject_GetAttrString(pkg, "size");
   Package::Size  result  = 0;

   if (PyLong_Check(sizeObj)) {
      result = PyLong_AsLongLong(sizeObj);
#if PY_VERSION_HEX < 0x03000000
   } else if (PyInt_Check(sizeObj)) {
      result = PyInt_AsLong(sizeObj);
#endif
   } else if (PY_IS_STRING(sizeObj)) {
      QString  str = pyString(sizeObj);
      errno = 0;
      result = strtoull(str.toStdString().c_str(), NULL, 10);
      if (errno != 0) {
         warning(tr("Internal error: package size not understood:"
                    " %1").arg(str));
      }
   } else if (PyCallable_Check(sizeObj)) {
      PyObject*  obj = pyMethodCall(pkg, "size", "()");

      if (PyLong_Check(obj)) {
         result = PyLong_AsLongLong(obj);
         Py_DECREF(obj);
#if PY_VERSION_HEX < 0x03000000
      } else if (PyInt_Check(obj)) {
         result = PyInt_AsLong(obj);
         Py_DECREF(obj);
#endif
      } else if (PY_IS_STRING(obj)) {
         QString  str = pyStringDec(obj);
         errno = 0;
         result = strtoull(str.toStdString().c_str(), NULL, 10);
         if (errno != 0) {
            warning(tr("Internal error: package size not understood:"
                       " %1").arg(str));
         }
      } else {
         warning(tr("Internal error: package size result not understood"));
         Py_XDECREF(obj);
      }
   } else {
      warning(tr("Internal error: package size not understood"));
   }
   
   Py_DECREF(sizeObj);

   return result;
}

QString  YumIO::bytesFormat(Package::Size  bytes)
{
   QString  result;
   double   units  = (double)bytes;
   QString  suffix = tr("b");

   if (bytes >= 1000ULL*1024*1024*1024*1024*1024) {
      units /= 1024.0*1024*1024*1024*1024*1024;
      suffix = tr("Eb");
   } else if (bytes >= 1000ULL*1024*1024*1024*1024) {
      units /= 1024.0*1024*1024*1024*1024;
      suffix = tr("Pb");
   } else if (bytes >= 1000ULL*1024*1024*1024) {
      units /= 1024.0*1024*1024*1024;
      suffix = tr("Tb");
   } else if (bytes >= 1000ULL*1024*1024) {
      units /= 1024.0*1024*1024;
      suffix = tr("Gb");
   } else if (bytes >= 1000ULL*1024) {
      units /= 1024.0*1024;
      suffix = tr("Mb");
   } else if (bytes >= 1000ULL) {
      units /= 1024.0;
      suffix = tr("Kb");
   }

   if (0.0 < units && units < 9.95) {
      result.sprintf("%.1f", units);
   } else {
      result.sprintf("%.0f", units);
   }
   result.append(suffix);

   return result;
}

time_t  YumIO::buildTimeOf(PyObject*  pkg)
{
   return pyTimeDec(PyObject_GetAttrString(pkg, "buildtime"));
}

#define TIME_1  "Mon Dec 31 "
#define TIME_2  "Mon Dec 31 03:00:00 "

char*  YumIO::dateFormat(time_t  secs)
{
   static char  buffer[1024];
   ctime_r(&secs, buffer);
   buffer[strlen(buffer)-1] = '\0'; // Strip the newline
   strcpy(&buffer[STRLEN(TIME_1)], &buffer[STRLEN(TIME_2)]);
   return buffer;
}

char*  YumIO::timeFormat(time_t  secs)
{
   static char  buffer[1024];
   ctime_r(&secs, buffer);
   buffer[strlen(buffer)-1] = '\0'; // Strip the newline
   return buffer;
}

QString  YumIO::yumVersionStr()
{
   if (!yumVerKnown) {
      if (pyLookup("yum", /* tolerant => */ true) != NULL) {
         PyObject*  ver = pyValue("yum", "__version__");
         if (ver == NULL) {
            ver = pyValue("cli", "__version__");
         }

         if (ver != NULL) {
            yumVer = pyString(ver);
         }
      }
      yumVerKnown = true;
   }

   return yumVer;
}

QString  YumIO::dnfVersionStr()
{
   if (!dnfVerKnown) {
      if (pyLookup("dnf", /* tolerant => */ true) != NULL) {
         PyObject*  ver = pyValue("dnf", "__version__");

         if (ver != NULL) {
            dnfVer = pyString(ver);
         }
      }
      dnfVerKnown = true;
   }

   return dnfVer;
}

long  YumIO::splitVersion(QString  str)
{
   QStringList  versions = str.split('.');
   long         result   = 0;

   int   i;
   long  mult;
   for (i = 0, mult = 10000; i < 3; i++, mult /= 100) {
      if (i >= versions.count()) break;

      errno = 0;
      long  value = strtoul(versions[i].toStdString().c_str(), NULL, 10);
      if (errno != 0) {
         static bool seen = false;
         if (!seen) {
            warning(tr("Internal error: yum/dnf version not understood: %1"
                       " errno=%2 (%3)")
                    .arg(str).arg(errno).arg(strerror(errno)));
            seen = true;
         }
      }

      result += value * mult;
   }

   return result;
}

long  YumIO::dnfVersion()
{
   QString  dnfVer = dnfVersionStr();
   if (!dnfVer.isEmpty()) {
      return splitVersion(dnfVer);
   }
   static bool seen = false;
   if (!seen) {
      warning(tr("Internal error: dnf version not found"));
      seen = true;
   }
   return 0;
}

bool  YumIO::hasDistribGroups()
{
   return true;
}

QString  YumIO::pyString(PyObject*  obj)
{
   if (obj == NULL) {
      // This can happen if the object was from an operation that raised
      // a python unhandled exception.  To avoid a cascade SIGSEGV, deal
      // with it by returning an empty string.
      warning(tr("Internal error: YumIO::pyString called with NULL"));
      return QString();
   } else if (PyUnicode_Check(obj)) {
      Py_ssize_t  size   = PyUnicode_GetSize(obj);
      wchar_t*    buffer = new wchar_t[size+1];
#if PY_VERSION_HEX >= 0x03000000
      Py_ssize_t  actual = PyUnicode_AsWideChar(obj, buffer, size+1);
#else
      Py_ssize_t  actual = PyUnicode_AsWideChar((PyUnicodeObject*)obj,
                                                buffer, size+1);
#endif
      if (actual == -1 && PyErr_Occurred()) handleException();
      if (actual >= 0) {
         buffer[actual] = '\0';
      }
      QString  result = QString::fromStdWString(buffer);
      delete [] buffer;
      return result;
#if PY_VERSION_HEX >= 0x03000000
   } else if (PyBytes_Check(obj)) {
      return QString(PyBytes_AsString(obj));
#endif
#if PY_VERSION_HEX < 0x03000000
   } else if (PyString_Check(obj)) {
      char*  str = Python::pyString(obj);
      QString  result(str);
      free(str);
      return result;
#endif
   } else if (obj == Py_None) {
      return QString();
   } else {
      warning(tr("Internal error: YumIO::pyString did not understand type: %1")
                 .arg(obj->ob_type->tp_name));
      return QString();
   }
}

QString  YumIO::pyStringDec(PyObject*  obj)
{
   QString  result = pyString(obj);
   Py_XDECREF(obj);
   return result;
}

// Generally the same as pyString.  But if a yum string is parenthesized
// (e.g. "(null)"), then it gets turned into a list.  This copes with that and
// turns it back into a string.  And to support pythonExceptionV, it also
// translates a variety of lists, tuples, and integers.
QString  YumIO::pyStringX (PyObject*  obj)
{
   if (PY_IS_STRING(obj)) {
      return pyString(obj);
   } else if (PyList_Check(obj)) {
      if (PyList_Size(obj) == 0) {
         return QString();
      } else {
         QString  result("(");
         PyObject*  it = PyObject_GetIter(obj);
         bool       first = true;
         for (;;) {
            PyObject*  item = PyIter_Next(it);
            if (item == NULL) break;
            if (!first) result.append(",\n ");
            result.append(pyStringX(item));
            Py_DECREF(item);
            first = false;
         }
         Py_DECREF(it);
         result.append(")");
         return result;
      }
   } else if (PyTuple_Check(obj)) {
      if (PyTuple_Size(obj) == 0) {
         return QString();
      } else {
         QString  result("(");
         PyObject*  it = PyObject_GetIter(obj);
         bool       first = true;
         for (;;) {
            PyObject*  item = PyIter_Next(it);
            if (item == NULL) break;
            if (!first) result.append(',');
            result.append(pyStringX(item));
            Py_DECREF(item);
            first = false;
         }
         Py_DECREF(it);
         result.append(")");
         return result;
      }
   } else if (PyLong_Check(obj)) {
      return QString::number(PyLong_AsLongLong(obj));
#if PY_VERSION_HEX < 0x03000000
   } else if (PyInt_Check(obj)) {
      return QString::number(PyInt_AsLong(obj));
#endif
   } else {
      warning(tr("Internal error: YumIO::pyStringX did not understand type: %1")
                 .arg(obj->ob_type->tp_name));
      return QString();
   }      
}

QString  YumIO::pyStringXDec (PyObject*  obj)
{
   QString  result = pyStringX(obj);
   Py_XDECREF(obj);
   return result;
}

QString  YumIO::pyMappingString (PyObject*    obj,
                                 const char*  attr)
{
   char*  copy = new char[strlen(attr)+1];
   strcpy(copy, attr);
   QString  result = pyStringDec(PyMapping_GetItemString(obj, copy));
   delete [] copy;
   return result;
}

time_t  YumIO::pyTime(PyObject*  obj)
{
   time_t  result = 0;

   if (PyLong_Check(obj)) {
      result = PyLong_AsLongLong(obj);
#if PY_VERSION_HEX < 0x03000000
   } else if (PyInt_Check(obj)) {
      result = PyInt_AsLong(obj);
#endif
   } else if (PY_IS_STRING(obj)) {
      QString  str = pyString(obj);
      errno = 0;
      result = strtoul(str.toStdString().c_str(), NULL, 10);
      if (errno != 0) {
         warning(tr("Internal error: YumIO::pyTimeDec did not understand:"
                    " %1").arg(str));
      }
   } else {
      warning(tr("Internal error: YumIO::pyTime did not understand type: %1")
                 .arg(obj->ob_type->tp_name));
   }

   return result;
}

time_t  YumIO::pyTimeDec(PyObject*  obj)
{
   time_t  result = 0;

   if (PyLong_Check(obj)) {
      result = PyLong_AsLongLong(obj);
      Py_DECREF(obj);
#if PY_VERSION_HEX < 0x03000000
   } else if (PyInt_Check(obj)) {
      result = PyInt_AsLong(obj);
      Py_DECREF(obj);
#endif
   } else if (PY_IS_STRING(obj)) {
      QString  str = pyStringDec(obj);
      errno = 0;
      result = strtoul(str.toStdString().c_str(), NULL, 10);
      if (errno != 0) {
         warning(tr("Internal error: YumIO::pyTimeDec did not understand:"
                    " %1").arg(str));
      }
   } else {
      warning(tr("Internal error: YumIO::pyTimeDec did not understand type: %1")
                 .arg(obj->ob_type->tp_name));
      Py_DECREF(obj);
   }

   return result;
}

QString  YumIO::pyDate(PyObject*  obj)
{
   QString  result;

   if (PyDate_Check(obj)) {
      int        year       = PyDateTime_GET_YEAR(obj);
      int        month      = PyDateTime_GET_MONTH(obj);
      int        day        = PyDateTime_GET_DAY(obj);
      PyObject*  weekdayObj = pyMethodCall(obj, "weekday", "()");
      int        weekday = 0;
      if (PyLong_Check(weekdayObj)) {
         weekday = PyLong_AsLongLong(weekdayObj);
      } else {
         warning(tr("Internal error: YumIO::pyDate did not understand"
                    " weekday type: %1")
                 .arg(obj->ob_type->tp_name));
      }
      Py_XDECREF(weekdayObj);
      if (month < 1 || month > 12) {
         warning(tr("Internal error: YumIO::pyDate found out-of-bounds"
                    " month: %1").arg(month));
         month = 0;
      }
      if (weekday < 0 || weekday > 6) {
         warning(tr("Internal error: YumIO::pyDate found out-of-bounds"
                    " weekday: %1").arg(weekday));
         weekday = 0;
      }
      static const char*  monthNames[]   = { "Jan", "Feb", "Mar",
                                             "Apr", "May", "Jun",
                                             "Jul", "Aug", "Sep",
                                             "Oct", "Nov", "Dec" };
      static const char*  weekdayNames[] = { "Mon", "Tue", "Wed",
                                             "Thu", "Fri", "Sat", "Sun" };
      result = QString("%1 %2 %3 %4").arg(weekdayNames[weekday])
                                     .arg(monthNames[month-1])
                                     .arg(day)
                                     .arg(year);
   } else {
      warning(tr("Internal error: YumIO::pyDate did not understand type: %1")
                 .arg(obj->ob_type->tp_name));
   }

   return result;
}

PyObject*  YumIO::pyDictGetItem (PyObject*       dict,
                                 const QString&  key)
{
   // This is because PyDict_GetItemString's 2nd parameter is a char* and not
   // a const char*.  It probably is just that its parameter profile isn't
   // const-correct.  But I'm not chancing it.
   char*      keyCopy = strdup(key.toStdString().c_str());
   PyObject*  result  = PyDict_GetItemString(dict, keyCopy);
   free(keyCopy);
   return result;
}

const char*  YumIO::pyDescribe (PyObject*  obj)
{
   if (PY_IS_STRING(obj)) {
      return pyString(obj).toStdString().c_str();
   } else if (PyLong_Check(obj)) {
      static char  buffer[1024];
      sprintf(buffer, "%lld", PyLong_AsLongLong(obj));
      return buffer;
#if PY_VERSION_HEX < 0x03000000
   } else if (PyInt_Check(obj)) {
      static char  buffer[1024];
      sprintf(buffer, "%ld", PyInt_AsLong(obj));
      return buffer;
#endif
   } else if (obj == Py_None) {
      return "*None*";
   } else {
      PyObject*    str    = pyMethodCall(obj, "__str__", "()");
      const char*  result = "*Obj-no-str*";
      if (PY_IS_STRING(str)) {
         static char  buffer[1024];
         sprintf(buffer, "%s(*Obj*)", pyString(str).toStdString().c_str());
         result = buffer;
      }
      Py_XDECREF(str);
      return result;
   }
}

bool  YumIO::reapChildren ()
{
   // If a previous apply was aborted somehow, there may still be applydeltarpm
   // children in zombie state.  Reap them so they are not discovered by the
   // next apply in yum.drpm.DeltaInfo._wait, where the exit of an unexpected
   // child can cause an assertion failure.

   for (;;) {
      pid_t  child = waitpid(-1, NULL, WNOHANG);
      if (child == 0)                     break;
      if (child == -1 && errno == ECHILD) break;
      if (child == -1) {
         error(tr("Unable to reap children; errno=%1 (%2)")
                  .arg(errno).arg(strerror(errno)));
         return false;
      }
      log(Log::WARNING, 0, tr("Reaped zombie child %1").arg(child));
   }

   return true;
}

void  YumIO::sigchldDisable ()
{
   struct sigaction  noAct;
   noAct.sa_handler = SIG_DFL;
   sigemptyset(&noAct.sa_mask);
   noAct.sa_flags   = 0;
   int  status = sigaction(SIGCHLD, &noAct, &savedSigChldAct);
   if (status == -1) {
      error(tr("Unable to disable SIGCHLD signal handler; errno=%1 (%2)")
               .arg(errno).arg(strerror(errno)));
   }
}

void  YumIO::sigchldRestore ()
{
   int  status = sigaction(SIGCHLD, &savedSigChldAct, NULL);
   if (status == -1) {
      error(tr("Unable to restore SIGCHLD signal handler; errno=%1 (%2)")
               .arg(errno)
               .arg(strerror(errno)));
   }
}

void  YumIO::commExceptSetup ()
{
   pyExceptExpect(xcpt.bdbQuit);
   pyExceptExpect(xcpt.yumBaseError);
   pyExceptExpect(xcpt.repoError);
   pyExceptExpect(PyExc_IOError);
   pyExceptExpect(PyExc_KeyboardInterrupt);
   pyExceptExpect(PyExc_AssertionError);
   pyExceptExpect(PyExc_RuntimeError);

   if (xcpt.sqlite3OperationalError != NULL) {
      pyExceptExpect(xcpt.sqlite3OperationalError);
   }
   if (xcpt.noMoreMirrorsRepoError != NULL) {
      pyExceptExpect(xcpt.noMoreMirrorsRepoError);
   }
}

bool  YumIO::commExceptRaised (PyObject*  result,
                               QString    descr)
{
   // Be careful here.  You must not do pyLookup() in this routine, as that can
   // call PyErr_Clear(), which makes the pyExceptMatches() queries all return
   // false.

   if (pyExceptMatches(result, PyExc_IOError)) {
      PyObject*  exception;
      PyObject*  v;
      PyObject*  tb;
      PyErr_Fetch(&exception, &v, &tb);
      PyErr_NormalizeException(&exception, &v, &tb);
      PyErr_Display(exception, v, tb);
      QString  value = pyStringDec(pyExceptValuef());
      error(tr("Unable to %1: I/O Error: %2\n\n"
               "Emitting a walkback to the Log window.")
               .arg(descr).arg(value));
      return true;
   } else if (pyExceptMatches(result, PyExc_KeyboardInterrupt)) {
      PyObject*  exception;
      PyObject*  v;
      PyObject*  tb;
      PyErr_Fetch(&exception, &v, &tb);
      PyErr_NormalizeException(&exception, &v, &tb);
      PyErr_Display(exception, v, tb);
      error(tr("Unable to %1: KeyboardInterrupt\n\n"
               "This usually means NUU was sent signal SIGINT,"
               " SIGTERM, or similar.\n\n"
               "But urlgrabber also can raise this error for several"
               " situations where it is just guessing.\n"
               "So, emitting a walkback to the Log window.")
               .arg(descr));
      return true;
   } else if (pyExceptMatches(result, PyExc_AssertionError)) {
      PyObject*  exception;
      PyObject*  v;
      PyObject*  tb;
      PyErr_Fetch(&exception, &v, &tb);
      PyErr_NormalizeException(&exception, &v, &tb);
      PyErr_Display(exception, v, tb);
      error(tr("Internal error: python assertion failed\n"
               "When reporting this issue, please include the contents of the\n"
               "Log window, which should contain a walkback."));
      return true;
   } else if (pyExceptMatches(result, PyExc_RuntimeError)) {
      QString  value = pyStringDec(pyExceptValuef());
      error(tr("Unable to %1: Runtime Error:\n\n%2").arg(descr).arg(value));
      return true;
   } else if (xcpt.noMoreMirrorsRepoError != NULL
              && pyExceptMatches(result, xcpt.noMoreMirrorsRepoError)) {
      QString  value = pyStringDec(pyExceptValuef());
      error(tr("Unable to %1: metadata unavailable from all repo mirrors:\n"
               "%2").arg(descr).arg(value));
      return true;
   } else if (pyExceptMatches(result, xcpt.repoError)
              || (xcpt.sqlite3OperationalError != NULL
                  && pyExceptMatches(result, xcpt.sqlite3OperationalError))) {
      QString  value = pyStringDec(pyExceptValuef());
      if (value.contains("database is locked")) {
         pythonException(tr("Unable to %1: a repository's sqlite database"
                            " was locked unexpectedly\n"
                            "On some distributions, PackageKit has been known"
                            " to cause this.").arg(descr));
         return true;
      } else {
         exceptionError();
         return true;
      }
   } else if (pyExceptMatches(result, xcpt.bdbQuit)
              || pyExceptMatches(result, xcpt.yumBaseError)) {
      pythonExceptionV(tr("Unable to %1").arg(descr));
      return true;
   }

   return false;
}

void  YumIO::pyExceptExpect (PyObject*  xcpt)
{
   logEntries.lock();
   logStartOp = logEntries.count();
   logEntries.unlock();
   Python::pyExceptExpect(xcpt);
}

void  YumIO::pyExceptClear ()
{
   logStartOp = -1;
   Python::pyExceptClear();
}

void  YumIO::pythonException (QString  text)
{
   int  startOp;
   if (logStartOp != -1) {
      startOp = logStartOp;
   } else {
      logEntries.lock();
      startOp = logEntries.count();
      logEntries.unlock();
   }

   log(Log::ERROR, 0, text);

   QString  msg;

   logEntries.lock();
   int  i;
   for (i = startOp; i < logEntries.count(); i++) {
      if (logEntries[i].kind == Log::ERROR && logEntries[i].value <= 2) {
         if (!msg.isEmpty()) {
            msg.append('\n');
         }
         msg.append(logEntries[i].text);
      }
   }
   logEntries.unlock();

   dialogs->scrollable(tr("NUU Yum Error"), msg);
}

void  YumIO::pythonExceptionV (QString  text)
{
   int  startOp;
   if (logStartOp != -1) {
      startOp = logStartOp;
   } else {
      logEntries.lock();
      startOp = logEntries.count();
      logEntries.unlock();
   }

   log(Log::ERROR, 0, pyStringXDec(pyExceptValuef()));
   log(Log::ERROR, 0, text);

   QString  msg;

   logEntries.lock();
   int  i;
   for (i = startOp; i < logEntries.count(); i++) {
      if (logEntries[i].kind == Log::ERROR && logEntries[i].value <= 2) {
         if (!msg.isEmpty()) {
            msg.append('\n');
         }
         msg.append(logEntries[i].text);
      }
   }
   logEntries.unlock();

   dialogs->scrollable(tr("NUU Yum Error"), msg);
}

void  YumIO::exceptionExpected()
{
#ifdef DEBUG_EXPECTED_EXCEPTIONS
   // This exception was totally expected.  But maybe for debugging I'd
   // want to know the particulars.  So, stash this at a very high log
   // level to imply its debugging nature.  Note that a Search operation
   // will cause tons of these.

   Log::Kind     oldStderrKind  = wr.stderrKind;
   unsigned int  oldStderrValue = wr.stderrValue;

   wr.stderrKind  = Log::NORMAL;
   wr.stderrValue = 10;

   PyObject*  exception;
   PyObject*  v;
   PyObject*  tb;
   PyErr_Fetch(&exception, &v, &tb);
   PyErr_NormalizeException(&exception, &v, &tb);
   PyErr_Display(exception, v, tb);
   PyErr_Restore(exception, v, tb);

   wr.stderrKind  = oldStderrKind;
   wr.stderrValue = oldStderrValue;

   Python::exceptionExpected();
#endif
}

/*
** Python type interfaces
*/

QString  YumIO::txmbrFormat(PyObject* txmbr)
{
   QString  name  = pyStringDec(PyObject_GetAttrString(txmbr, "name"));
   QString  arch  = pyStringDec(PyObject_GetAttrString(txmbr, "arch"));
   long     epoch = PyLong_AsLongLong(PyObject_GetAttrString(txmbr, "epoch"));
   QString  ver   = pyStringDec(PyObject_GetAttrString(txmbr, "version"));
   QString  rel   = pyStringDec(PyObject_GetAttrString(txmbr, "release"));

   if (epoch == 0) {
      return tr("%1 %3-%4 (%2)").arg(name).arg(arch).arg(ver).arg(rel);
   } else {
      return tr("%1 %3:%4-%5 (%2)")
                .arg(name).arg(arch).arg(epoch).arg(ver).arg(rel);
   }
}

QString  YumIO::txmbrRepoFormat(PyObject*  txmbr)
{
   QString  repoid = pyStringDec(PyObject_GetAttrString(txmbr, "from_repo"));

   return txmbrFormat(txmbr).append(tr(" [%1]").arg(repoid));
}

int YumIO::showTransPkgs(QString    action,
                         PyObject*  txmbrs)
{
   PyObject*  it = PyObject_GetIter(txmbrs);
   for (;;) {
      PyObject*  txmbr = PyIter_Next(it);
      if (txmbr == NULL) break;
      ts.descr.append(tr("\t%1: %2\n").arg(action).arg(txmbrFormat(txmbr)));
      Py_DECREF(txmbr);
   }

   Py_DECREF(it);

   return 1;
}

int YumIO::showTransPkgsRepo(QString    action,
                             PyObject*  txmbrs)
{
   PyObject*  it = PyObject_GetIter(txmbrs);
   for (;;) {
      PyObject*  txmbr = PyIter_Next(it);
      if (txmbr == NULL) break;
      ts.descr.append(tr("\t%1: %2\n").arg(action)
                                      .arg(txmbrRepoFormat(txmbr)));
      Py_DECREF(txmbr);
   }

   Py_DECREF(it);

   return 1;
}

PyObject*  YumIO::findRelation(PyObject*    relatedto,
                               const char*  relation)
{
   PyObject*  it = PyObject_GetIter(relatedto);
   for (;;) {
      PyObject*  tuple = PyIter_Next(it);
      if (tuple == NULL) break;

      if (Python::pyStringEqual(PyTuple_GetItem(tuple, 1), relation)) {
         PyObject*  result = PyTuple_GetItem(tuple, 0);
         // Make result a new reference
         Py_INCREF(result);
         Py_DECREF(tuple);
         Py_DECREF(it);
         return result;
      }

      Py_DECREF(tuple);
   }

   Py_DECREF(it);

   return NULL;
}

void  YumIO::noteTransPkgs(Package::Action  action,
                           PyObject*        txmbrs)
{
   PyObject*  it = PyObject_GetIter(txmbrs);

   for (;;) {
      PyObject*  txmbr = PyIter_Next(it);
      if (txmbr == NULL) break;

      QString  name  = pyStringDec(PyObject_GetAttrString(txmbr, "name"));
      QString  arch  = pyStringDec(PyObject_GetAttrString(txmbr, "arch"));
      Package::Entry*  entry = hash[EntryKey(false, name, getBiArch(arch))];
      switch (action) {
      case Package::INSTALL:
         ts.run.nominalGrabSize += PG_R_GRAB(entry->size);
         if (!QFile::exists(getLocalPkg(entry->obj))) {
            ts.run.totalGrabSize += PG_R_GRAB(entry->size);
         }
         ts.run.totalTsSize += PG_R_INSTALL(entry->size);
         if (entry->apply.action == Package::NONE) {
            entry->apply.action = Package::INSTALL;
         }
         break;
      case Package::UPDATE:
         ts.run.nominalGrabSize += PG_R_GRAB(entry->size);
         if (!QFile::exists(getLocalPkg(entry->obj))) {
            ts.run.totalGrabSize += PG_R_GRAB(entry->size);
         }
         ts.run.totalTsSize += PG_R_UPDATE(entry->size);
         if (entry->apply.action == Package::NONE) {
            entry->apply.action = Package::UPDATE;
         }
         break;
      case Package::REMOVE:
         {
            QString  verrel = getVerRelTxmbr(txmbr);
            Package::Entry*  e = entry;
            if (e->state == Package::OUT_OF_DATE) {
               // Skip the updateable package, since it isn't actually
               // installed currently.
               e = e->next;
            }
            for (; e != NULL; e = e->next) {
               if (e->verrel == verrel) {
                  ts.run.totalTsSize += PG_R_REMOVE(e->size);
                  break;
               }
            }
         }
         if (entry->apply.action == Package::NONE) {
            entry->apply.action = Package::REMOVE;
         }
         break;
      default:
         // DO NOTHING
         break;
      }

      Py_DECREF(txmbr);
   }

   Py_DECREF(it);
}

void  YumIO::error(QString  text)
{
   log(Log::ERROR, 0, text);
   dialogs->critical(tr("NUU Yum Error"), text);
}

void  YumIO::warning(QString  text)
{
   log(Log::WARNING, 0, text);
   dialogs->warning(tr("NUU Yum Warning"), text);
}

void  YumIO::error(const char*  text)
{
   error(QString(text));
}


void  YumIO::warning(const char*  text)
{
   warning(QString(text));
}
