#
# Debugging:
#
# To use the pdb debugger on code in this file, you will need to disable
# the parts of NUU that capture output for the Log window.  Do that in
# two places.  In YumIO.c, change:
#    #undef DEBUG_DISABLE_CAPTURERS
#       =>
#    #define DEBUG_DISABLE_CAPTURERS
# Then just insert the following wherever you need to gain control with pdb:
#    debug()
# If stdin, stdout & stderr still give you troubles, you also could change this
# at the end of this file:
#    #sys.stdout = YumIOStdout()
#    #sys.stderr = YumIOStderr()
#    #sys.stdin = YumIOStdin()
#

import sys
import os
sys.path.insert(0, '%s/python' % os.getenv('NUUSHARE', '/usr/share/nuu'))

from pprint import pprint

def debug():
    import sys
    old_stdout = sys.stdout
    sys.stdout = sys.__stdout__
    import pdb
    pdb.set_trace()
    sys.stdout = old_stdout

# If the garbage collector is enabled (the default), if anyone does an
# update, presses Cancel, does it again, and presses OK, the update will
# complete successfully, but then the re-initialization will fail with an
# assertion:
#    Modules/gcmodule.c:378: delete_garbage:
#    Assertion `((((PyGC_Head *)(op)-1))->gc.gc_refs >= 0)' failed.
# I went so far as disabling every single Py_DECREF in the C++ code and
# verified that there were Py_INCREF's before the calls to PyTuple_SetItem
# and PyList_SetItem, just to make sure I wasn't leaving any dangling
# references, but the gc still blew up.  I suspect this is a bug in Python's
# gc, so I'm disabling it.
import gc
gc.disable()

# For versions of Python that do GIL locking, NUU must act as a multi-threaded
# application (even though NUU itself guarantees that python cannot be running
# in more than one thread at a time).  But the YumBaseCli.doTransaction will
# attempt to call signal.signal, and that is not allowed in any non-main thread.
# To workaround this, create a new YumIOSignal that mostly works like the
# original signal module, but where the signal function is overridden to do no
# real work.
import types
import signal
__signal__ = signal
YumIOSignal = types.ModuleType('YumIOSignal', 'YumIO replacement for signal')
for meth in dir(__signal__):
    if meth not in [ '__name__', '__doc__', '__file__' ]:
        setattr(YumIOSignal, meth, getattr(__signal__, meth))
YumIOSignal.signal = lambda num, obj: __signal__.getsignal(num)
sys.modules['signal'] = YumIOSignal
import signal

#
# Main body of YumIO
#

import locale
import dnf.cli
import yumIOC
import dnf.yum
import dnf.rpm
import rpm
import distutils.version

def dnfVersion():
    if hasattr(dnf, '__version__'):
        return distutils.version.StrictVersion(dnf.yum.__version__)
    else:
        return distutils.version.StrictVersion('0.0')

def hasDistribGroups():
    return True

def usesPythonLogging():
    return True

def hasUpdatedTransactions():
    return True

# Used to override sys.stdin so that it will claim that it isatty, which
# convinces a yum check that it is not running unattended.
__stdin__ = sys.stdin
class YumIOStdin:
    def __init__(self):
        pass
    def isatty(self):
        return 1
for meth in dir(__stdin__):
    if not hasattr(YumIOStdin, meth):
        setattr(YumIOStdin, meth, getattr(__stdin__, meth))

# Used to override sys.stdout
class YumIOStdout:
    def write(self, msg):
        yumIOC.stdoutWrite(msg)

    def isatty(self):
        return 0

    def flush(self):
        pass

# Used to override sys.stderr
class YumIOStderr:
    def write(self, msg):
        yumIOC.stderrWrite(msg)

    def isatty(self):
        return 0

    def flush(self):
        pass

    # rpmtrans calls rpm.setLogFile(sys.stderr), and that demands a stderr with
    # a working PyObject_AsFileDescriptor.  So, provide one.
    def fileno(self):
        return sys.__stderr__.fileno()

# Used to override BaseCli's userconfirm
class YumIOOutput(dnf.cli.output.Output):
    def userconfirm(self, msg=None, defaultyes_msg=None):
        sys.stdout.flush()
        sys.stderr.flush()
        return yumIOC.userconfirm()

    def reportDownloadSize(self, packages, install_only=False):
        yumIOC.reportDownloadSize(packages)
        return super(YumIOOutput, self).reportDownloadSize(packages,
                                                           install_only)

    def reportRemoveSize(self, packages):
        yumIOC.reportRemoveSize(packages)
        return super(YumIOOutput, self).reportRemoveSize(packages)

    def list_transaction(self, transaction):
        yumIOC.listTransaction(self, transaction)
        return super(YumIOOutput, self).list_transaction(transaction)

    def post_transaction_output(self, transaction):
        yumIOC.postTransactionOutput(self, transaction)
        return super(YumIOOutput, self).post_transaction_output(transaction)

# Used to replace the dnf.cli.progress.MultiFileProgressMeter, overriding each
# repo's progress bar (via set_progress_bar).
class YumIOMultiFileProgressMeter(dnf.callback.DownloadProgress):
    def start(self, total_files, total_size, total_drpms=0):
        return yumIOC.downloadStart(total_files, total_size, total_drpms)

    def progress(self, payload, done):
        return yumIOC.downloadProgress(payload, done)

    def message(self, msg):
        return yumIOC.downloadMessage(total_files, total_size)

    def end(self, payload, status, msg):
        return yumIOC.downloadEnd(payload, status, msg)

# YumIOTransactionDisplay replaces the dnf.cli.output.CliTransactionDisplay,
# which is passed as part of the display list in dnf.cli.BaseCli.do_transaction.
#
# However, that do_transaction does a LOT and it would be problematic to
# replicate it.  So I need to trick it into creating my own TransactionDisplay
# when it wants to create its own dnf.cli.output.CliTransactionDisplay.  Do this
# by breaking into the dnf.cli.output module and replacing the class there with
# my own.
class YumIOTransactionDisplay(dnf.yum.rpmtrans.LoggingTransactionDisplay):
    def __init__(self):
        super(YumIOTransactionDisplay, self).__init__()

    def progress(self, package, action, ti_done, ti_total, ts_done, ts_total):
        return yumIOC.rpmtransProgress(package, action,
                                       ti_done, ti_total,
                                       ts_done, ts_total)

    def scriptout(self, msgs):
        return yumIOC.rpmtransScriptout(msgs)
dnf.cli.output.CliTransactionDisplay = YumIOTransactionDisplay
        
# YumIORPMTransaction replaces the dnf.yum.rpmtrans.RPMTransaction, which is
# created twice in dnf.Base.do_transaction, once for the transaction test
# (a.k.a. "test"), and again for the real transaction (a.k.a. "run").
#
# However, that do_transaction does a LOT and it would be problematic to
# replicate it.  So I need to trick it into creating my own RPMTransaction when
# it wants to create its own dnf.yum.rpmtrans.RPMTransaction.  Do this by
# breaking into the dnf.yum.rpmtrans module and replacing the class there with
# my own.
#
# This class goes beyond the callbacks defined in the TransactionDisplay above
# by also making callbacks when starting and ending a transaction test or
# transaction run.  It detects that by intercepting the __init__ and __del__
# methods.  Alternatively, it could've intercepted the callback method and
# checked the "what" for RPMCALLBACK_TRANS_START or RPMCALLBACK_TRANS_STOP, but
# that's more complicated and more likely to break because of later dnf changes.
class YumIORPMTransaction(dnf.yum.rpmtrans.RPMTransaction):
    def __init__(self, base, test=False, displays=()):
        yumIOC.rpmtransInit(test)
        super(YumIORPMTransaction, self).__init__(base,
                                                  test=test, displays=displays)

    def __del__(self):
        yumIOC.rpmtransDel()
        super(YumIORPMTransaction, self).__del__()
dnf.yum.rpmtrans.RPMTransaction = YumIORPMTransaction

class YumBaseCli(dnf.cli.cli.BaseCli):
    def gpgsigcheck(self, pkgs):
        yumIOC.gpgsigcheck()
        return super(YumBaseCli, self).gpgsigcheck(pkgs)

class YumIO(dnf.cli.cli.Cli):
    def __init__(self):
        locale.setlocale(locale.LC_ALL, '')
        basecli = YumBaseCli(dnf.conf.Conf())
        dnf.cli.cli.Cli.__init__(self, basecli)
        dnf.cli.cli.Cli.configure(self, [ 'makecache' ])
        bar = YumIOMultiFileProgressMeter()
        self.base.repos.all().set_progress_bar(bar)
        self.base.output = YumIOOutput(self.base, self.base.conf)
        self.base.output.progress = bar

    # Utility functions

    def log(self, value, msg):
        return yumIOC.log(value, msg)

    def errorlog(self, value, msg):
        return yumIOC.errorlog(value, msg)

    def lock(self):
        pass

    def unlock(self):
        pass

    def enableRepo(self, id):
        repo = self.base.repos.get_matching(id)
        repo.enable()

    def disableRepo(self, id):
        repo = self.base.repos.get_matching(id)
        repo.disable()
    
    def setupRepos(self):
        try:
            self.log(2, 'Setting up Repos')
            # It provide impossible to load 'otherdata' metadata on demand after
            # initialization.  So instead, ensure it is loaded during
            # initialization regardless of whether or not it will be needed.
            for repo in self.base.repos.iter_enabled():
                repo.load_metadata_other = True
            self.base.fill_sack(load_system_repo='auto',
                                load_available_repos=True)
            return None
        except:
            raise

    def setupTs(self):
        #return self.doTsSetup()
        pass
        
    def setupTsGroup(self):
        #return self.doGroupSetup()
        pass

    # High-level functions called from YumIO.c

    def collectUpdates(self):
        return self.base.returnPkgLists('upgrades').updates

    def collectAll(self):
        return self.base.returnPkgLists('all')

    def collectGroupInfo(self):
        self.base.read_comps(arch_filter=True)
        return self.base.comps.groups

    def collectDone(self):
        self.base._closeRpmDB()
        #self.doTsSetup()

    def apply(self, installs, updates, removes):
        try:
            for pkg in installs:
                self.base.package_install(pkg)
            for pair in updates:
                self.base.package_upgrade(pair[0])
            for pkg in removes:
                self.base.package_remove(pkg)
            result = self.base.resolve()
            if not result:
                raise Error.YumBaseError('Nothing to do')
            self.base.do_transaction()
            # If the transaction succeeds, we need to re-load the sack.
            self.base.reset(goal=True, sack=True)
        except:
            self.base.reset(goal=True)
            if hasattr(self, 'history'):
                self.history.close()
            self.base._closeRpmDB()
            raise

    def applyFailCleanup(self):
        pass

    def collectFileLists(self):
        pass

    def collectOtherData(self):
        pass

    def flush(self):
        pass

    def getBiArch(self, arch):
        try:
            return dnf.rpm.basearch(arch)
        except KeyError:
            return arch

    def isBetterThan(self, pkg1, pkg2):
        return True

#sys.stdout = YumIOStdout()
#sys.stderr = YumIOStderr()
#sys.stdin = YumIOStdin()
